import os
import json
import unittest
import jc.parsers.git_log_s
from jc.exceptions import ParseError

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

# To create streaming output use:
# $ cat git-log.out | jc --git-log-s | jello -c > git-log-streaming.json    # or jq -sc


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log.out'), 'r', encoding='utf-8') as f:
        generic_git_log = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-short.out'), 'r', encoding='utf-8') as f:
        generic_git_log_short = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-short-stat.out'), 'r', encoding='utf-8') as f:
        generic_git_log_short_stat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-short-shortstat.out'), 'r', encoding='utf-8') as f:
        generic_git_log_short_shortstat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-medium.out'), 'r', encoding='utf-8') as f:
        generic_git_log_medium = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-medium-stat.out'), 'r', encoding='utf-8') as f:
        generic_git_log_medium_stat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-medium-shortstat.out'), 'r', encoding='utf-8') as f:
        generic_git_log_medium_shortstat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-full.out'), 'r', encoding='utf-8') as f:
        generic_git_log_full = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-full-stat.out'), 'r', encoding='utf-8') as f:
        generic_git_log_full_stat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-full-shortstat.out'), 'r', encoding='utf-8') as f:
        generic_git_log_full_shortstat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-fuller.out'), 'r', encoding='utf-8') as f:
        generic_git_log_fuller = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-fuller-stat.out'), 'r', encoding='utf-8') as f:
        generic_git_log_fuller_stat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-fuller-shortstat.out'), 'r', encoding='utf-8') as f:
        generic_git_log_fuller_shortstat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-oneline.out'), 'r', encoding='utf-8') as f:
        generic_git_log_oneline = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-oneline-stat.out'), 'r', encoding='utf-8') as f:
        generic_git_log_oneline_stat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-oneline-shortstat.out'), 'r', encoding='utf-8') as f:
        generic_git_log_oneline_shortstat = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-hash-in-message-fix.out'), 'r', encoding='utf-8') as f:
        generic_git_log_fuller_hash_in_message_fix = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-is-hash-regex-fix.out'), 'r', encoding='utf-8') as f:
        generic_git_log_fuller_is_hash_regex_fix = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-blank-author-fix.out'), 'r', encoding='utf-8') as f:
        generic_git_log_blank_author_fix = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-short-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_short_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-short-stat-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_short_stat_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-short-shortstat-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_short_shortstat_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-medium-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_medium_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-medium-stat-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_medium_stat_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-medium-shortstat-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_medium_shortstat_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-full-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_full_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-full-stat-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_full_stat_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-full-shortstat-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_full_shortstat_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-fuller-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_fuller_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-fuller-stat-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_fuller_stat_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-fuller-shortstat-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_fuller_shortstat_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-oneline-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_oneline_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-oneline-stat-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_oneline_stat_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-oneline-shortstat-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_oneline_shortstat_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-streaming-ignore-exceptions.json'), 'r', encoding='utf-8') as f:
        generic_git_log_streaming_ignore_exceptions_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-hash-in-message-fix-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_fuller_hash_in_message_fix_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-is-hash-regex-fix-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_fuller_is_hash_regex_fix_streaming_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/git-log-blank-author-fix-streaming.json'), 'r', encoding='utf-8') as f:
        generic_git_log_blank_author_fix_streaming_json = json.loads(f.read())


    def test_git_log_s_nodata(self):
        """
        Test 'git_log' with no data
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse([], quiet=True)), [])

    def test_git_log_s_unparsable(self):
        data = 'unparsable data'
        g = jc.parsers.git_log_s.parse(data.splitlines(), quiet=True)
        with self.assertRaises(ParseError):
            list(g)

    def test_git_log_s_ignore_exceptions_success(self):
        """
        Test 'git log' with -qq (ignore_exceptions) option
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log.splitlines(), quiet=True, ignore_exceptions=True)), self.generic_git_log_streaming_ignore_exceptions_json)

    def test_ping_s_ignore_exceptions_error(self):
        """
        Test 'ping' with -qq (ignore_exceptions) option option and error
        """
        data_in = 'not git log'
        expected = json.loads('[{"_jc_meta":{"success":false,"error":"ParseError: Not git_log_s data","line":"not git log"}}]')
        self.assertEqual(list(jc.parsers.git_log_s.parse(data_in.splitlines(), quiet=True, ignore_exceptions=True)), expected)

    def test_git_log_s(self):
        """
        Test 'git_log'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log.splitlines(), quiet=True)), self.generic_git_log_streaming_json)

    def test_git_log_short_s(self):
        """
        Test 'git_log --format=short'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_short.splitlines(), quiet=True)), self.generic_git_log_short_streaming_json)

    def test_git_log_short_stat_s(self):
        """
        Test 'git_log --format=short --stat'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_short_stat.splitlines(), quiet=True)), self.generic_git_log_short_stat_streaming_json)

    def test_git_log_short_shortstat_s(self):
        """
        Test 'git_log --format=short --shortstat'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_short_shortstat.splitlines(), quiet=True)), self.generic_git_log_short_shortstat_streaming_json)

    def test_git_log_medium_s(self):
        """
        Test 'git_log --format=medium'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_medium.splitlines(), quiet=True)), self.generic_git_log_medium_streaming_json)

    def test_git_log_medium_stat_s(self):
        """
        Test 'git_log --format=medium --stat'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_medium_stat.splitlines(), quiet=True)), self.generic_git_log_medium_stat_streaming_json)

    def test_git_log_medium_shortstat_s(self):
        """
        Test 'git_log --format=medium --shortstat'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_medium_shortstat.splitlines(), quiet=True)), self.generic_git_log_medium_shortstat_streaming_json)

    def test_git_log_full_s(self):
        """
        Test 'git_log --format=full'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_full.splitlines(), quiet=True)), self.generic_git_log_full_streaming_json)

    def test_git_log_full_stat_s(self):
        """
        Test 'git_log --format=full --stat'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_full_stat.splitlines(), quiet=True)), self.generic_git_log_full_stat_streaming_json)

    def test_git_log_full_shortstat_s(self):
        """
        Test 'git_log --format=full --shortstat'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_full_shortstat.splitlines(), quiet=True)), self.generic_git_log_full_shortstat_streaming_json)

    def test_git_log_fuller_s(self):
        """
        Test 'git_log --format=fuller'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_fuller.splitlines(), quiet=True)), self.generic_git_log_fuller_streaming_json)

    def test_git_log_fuller_stat_s(self):
        """
        Test 'git_log --format=fuller --stat'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_fuller_stat.splitlines(), quiet=True)), self.generic_git_log_fuller_stat_streaming_json)

    def test_git_log_fuller_shortstat_s(self):
        """
        Test 'git_log --format=fuller --shortstat'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_fuller_shortstat.splitlines(), quiet=True)), self.generic_git_log_fuller_shortstat_streaming_json)

    def test_git_log_oneline_s(self):
        """
        Test 'git_log --format=oneline'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_oneline.splitlines(), quiet=True)), self.generic_git_log_oneline_streaming_json)

    def test_git_log_oneline_stat_s(self):
        """
        Test 'git_log --format=oneline --stat'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_oneline_stat.splitlines(), quiet=True)), self.generic_git_log_oneline_stat_streaming_json)

    def test_git_log_oneline_shortstat_s(self):
        """
        Test 'git_log --format=oneline --shortstat'
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_oneline_shortstat.splitlines(), quiet=True)), self.generic_git_log_oneline_shortstat_streaming_json)

    def test_git_log_fuller_hash_in_message_fix(self):
        """
        Test 'git_log --format=fuller --stat' fix for when a commit message
        contains a line that is only a commit hash value.
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_fuller_hash_in_message_fix.splitlines(), quiet=True)), self.generic_git_log_fuller_hash_in_message_fix_streaming_json)

    def test_git_log_fuller_is_hash_fix(self):
        """
        Test 'git_log --format=fuller --stat' fix for when a commit message
        contains a line that evaluated as true to an older _is_hash regex
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_fuller_is_hash_regex_fix.splitlines(), quiet=True)), self.generic_git_log_fuller_is_hash_regex_fix_streaming_json)

    def test_git_log_blank_author_fix(self):
        """
        Test 'git_log' fix for when a commit author has a blank name,
        empty email, or both
        """
        self.assertEqual(list(jc.parsers.git_log_s.parse(self.generic_git_log_blank_author_fix.splitlines(), quiet=True)), self.generic_git_log_blank_author_fix_streaming_json)


if __name__ == '__main__':
    unittest.main()



r"""jc - JSON Convert `amixer sget` command output parser

Usage (cli):

    $ amixer sget <control_name> | jc --amixer
    $ amixer sget Master | jc --amixer
    $ amixer sget Capture | jc --amixer
    $ amixer sget Speakers | jc --amixer

Usage (module):

    import jc
    result = jc.parse('amixer', <amixer_sget_command_output>)

Schema:

    {
        "control_name":                     string,
        "capabilities": [
                                            string
        ],
        "playback_channels": [
                                            string
        ],
        "limits": {
            "playback_min":                 integer,
            "playback_max":                 integer
        },
        "mono": {
            "playback_value":               integer,
            "percentage":                   integer,
            "db":                           float,
            "status":                       boolean
        }
    }

Examples:

    $ amixer sget Master | jc --amixer -p
    {
      "control_name": "Capture",
      "capabilities": [
        "cvolume",
        "cswitch"
      ],
      "playback_channels": [],
      "limits": {
        "playback_min": 0,
        "playback_max": 63
      },
      "front_left": {
        "playback_value": 63,
        "percentage": 100,
        "db": 30.0,
        "status": true
      },
      "front_right": {
        "playback_value": 63,
        "percentage": 100,
        "db": 30.0,
        "status": true
      }
    }

    $ amixer sget Master | jc --amixer -p -r
    {
        "control_name": "Master",
        "capabilities": [
            "pvolume",
            "pvolume-joined",
            "pswitch",
            "pswitch-joined"
        ],
        "playback_channels": [
            "Mono"
        ],
        "limits": {
            "playback_min": "0",
            "playback_max": "87"
        },
        "mono": {
            "playback_value": "87",
            "percentage": "100%",
            "db": "0.00db",
            "status": "on"
        }
    }
"""
from typing import Dict

import jc.utils
from jc.utils import convert_to_int
from jc.exceptions import ParseError

class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`amixer` command parser'
    author = 'Eden Refael'
    author_email = 'edenraf@hotmail.com'
    compatible = ['linux']
    magic_commands = ['amixer']
    tags = ['command']


__version__ = info.version


def _process(proc_data: Dict) -> Dict:
    """
    Processes raw structured data to match the schema requirements.

    Parameters:
        proc_data: (dict) raw structured data from the parser

    Returns:
        (dict) processed structured data adhering to the schema
    """
    if not proc_data:
        return {}

    # Initialize the processed dictionary
    processed = {
        "control_name": proc_data.get("control_name", ""),
        "capabilities": proc_data.get("capabilities", []),
        "playback_channels": proc_data.get("playback_channels", []),
        "limits": {
            "playback_min": convert_to_int(proc_data.get("limits", {}).get("playback_min", 0)),
            "playback_max": convert_to_int(proc_data.get("limits", {}).get("playback_max", 0)),
        },
    }

    # Process Mono or channel-specific data
    channels = ["mono", "front_left", "front_right"]
    for channel in channels:
        if channel in proc_data:
            channel_data = proc_data[channel]
            processed[channel] = {
                "playback_value": convert_to_int(channel_data.get("playback_value", 0)),
                "percentage": convert_to_int(channel_data.get("percentage", "0%").strip("%")),
                "db": float(channel_data.get("db", "0.0db").strip("db")),
                "status": channel_data.get("status", "off") == "on",
            }

    return processed


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> Dict:
    """
    Main text parsing function

    Parameters:
        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:
        Dictionary. Raw or processed structured data.
    """
    """
    The Algorithm for parsing the `amixer sget` command, Input Explained/Rules/Pseudo Algorithm:
    1. There will always be the first line which tells the user about the control name.
    2. There will always be the Capabilities which include many of capabilities - It will be listed and separated by `" "`.
    3. After that we'll need to distinct between the Channel - Could be many of channels - It will be listed and separated
       by `" "`.
        3a. Capture channels - List of channels
        3b. Playback channels - List of channels
    4. Limits - We'll always have the minimum limit and the maximum limit.

    Input Example:
        1."":~$ amixer sget Capture
        Simple mixer control 'Capture',0
          Capabilities: cvolume cswitch
          Capture channels: Front Left - Front Right
          Limits: Capture 0 - 63
          Front Left: Capture 63 [100%] [30.00db] [on]
          Front Right: Capture 63 [100%] [30.00db] [on]

        2."":~$ amixer sget Master
        Simple mixer control 'Master',0
          Capabilities: pvolume pvolume-joined pswitch pswitch-joined
          Playback channels: Mono
          Limits: Playback 0 - 87
          Mono: Playback 87 [100%] [0.00db] [on]

        3."":~$ amixer sget Speaker
        Simple mixer control 'Speaker',0
          Capabilities: pvolume pswitch
          Playback channels: Front Left - Front Right
          Limits: Playback 0 - 87
          Mono:
          Front Left: Playback 87 [100%] [0.00db] [on]
          Front Right: Playback 87 [100%] [0.00db] [on]

        4."":~$ amixer sget Headphone
        Simple mixer control 'Headphone',0
          Capabilities: pvolume pswitch
          Playback channels: Front Left - Front Right
          Limits: Playback 0 - 87
          Mono:
          Front Left: Playback 0 [0%] [-65.25db] [off]
          Front Right: Playback 0 [0%] [-65.25db] [off]
    """
    # checks os compatibility and print a stderr massage if not compatible. quiet True could remove this check.
    jc.utils.compatibility(__name__, info.compatible, quiet)

    # check if string
    jc.utils.input_type_check(data)

    # starts the parsing from here
    mapping: Dict = {}

    if jc.utils.has_data(data):
        # split lines and than work on each line
        lines = data.splitlines()
        first_line = lines[0].strip()

        # Extract the control name from the first line
        if first_line.startswith("Simple mixer control"):
            control_name = first_line.split("'")[1]
        else:
            raise ParseError("Invalid amixer output format: missing control name.")
        # map the control name
        mapping["control_name"] = control_name

        # Process subsequent lines for capabilities, channels, limits, and channel-specific mapping.
        # gets the lines from the next line - because we already took care the first line.
        for line in lines[1:]:
            # strip the line (maybe there are white spaces in the begin&end)
            line = line.strip()

            if line.startswith("Capabilities:"):
                mapping["capabilities"] = line.split(":")[1].strip().split()
            elif line.startswith("Playback channels:"):
                mapping["playback_channels"] = line.split(":")[1].strip().split(" - ")
            elif line.startswith("Limits:"):
                limits = line.split(":")[1].strip().split(" - ")
                mapping["limits"] = {
                    "playback_min": limits[0].split()[1],
                    "playback_max": limits[1]
                }
            elif line.startswith("Mono:") or line.startswith("Front Left:") or line.startswith("Front Right:"):
                # Identify the channel name and parse its information
                channel_name = line.split(":")[0].strip().lower().replace(" ", "_")
                channel_info = line.split(":")[1].strip()
                # Example: "Playback 255 [100%] [0.00db] [on]"
                channel_data = channel_info.split(" ")
                if channel_data[0] == "":
                    continue

                if "dB" in channel_data[3]:
                    db_value = channel_data[3].strip("[]")
                    status = channel_data[4].strip("[]")
                else:
                    db_value = "0.0db"
                    status = channel_data[3].strip("[]")

                playback_value = channel_data[1]
                percentage = channel_data[2].strip("[]")  # Extract percentage e.g., "100%"

                # Store channel mapping in the dictionary
                mapping[channel_name] = {
                    "playback_value": playback_value,
                    "percentage": percentage,
                    "db": db_value.lower(),
                    "status": status
                }

    return mapping if raw else _process(mapping)



r"""jc - JSON Convert `dir` command output parser

Options supported:
- `/T timefield`
- `/O sortorder`
- `/C, /-C`
- `/S`

The "Magic" syntax is not supported since the `dir` command is a shell
builtin.

The `epoch` calculated timestamp field is naive. (i.e. based on the local
time of the system the parser is run on)

Usage (cli):

    C:\> dir | jc --dir

Usage (module):

    import jc
    result = jc.parse('dir', dir_command_output)

Schema:

    [
      {
        "date":         string,
        "time":         string,
        "epoch":        integer,    # naive timestamp
        "dir":          boolean,
        "size":         integer,
        "filename:      string,
        "parent":       string
      }
    ]

Examples:

    C:\> dir | jc --dir -p
    [
      {
        "date": "03/24/2021",
        "time": "03:15 PM",
        "dir": true,
        "size": null,
        "filename": ".",
        "parent": "C:\\Program Files\\Internet Explorer",
        "epoch": 1616624100
      },
      {
        "date": "03/24/2021",
        "time": "03:15 PM",
        "dir": true,
        "size": null,
        "filename": "..",
        "parent": "C:\\Program Files\\Internet Explorer",
        "epoch": 1616624100
      },
      {
        "date": "12/07/2019",
        "time": "02:49 AM",
        "dir": true,
        "size": null,
        "filename": "en-US",
        "parent": "C:\\Program Files\\Internet Explorer",
        "epoch": 1575715740
      },
      {
        "date": "12/07/2019",
        "time": "02:09 AM",
        "dir": false,
        "size": 54784,
        "filename": "ExtExport.exe",
        "parent": "C:\\Program Files\\Internet Explorer",
        "epoch": 1575713340
      },
      ...
    ]

    C:\> dir | jc --dir -p -r
    [
      {
        "date": "03/24/2021",
        "time": "03:15 PM",
        "dir": true,
        "size": null,
        "filename": ".",
        "parent": "C:\\Program Files\\Internet Explorer"
      },
      {
        "date": "03/24/2021",
        "time": "03:15 PM",
        "dir": true,
        "size": null,
        "filename": "..",
        "parent": "C:\\Program Files\\Internet Explorer"
      },
      {
        "date": "12/07/2019",
        "time": "02:49 AM",
        "dir": true,
        "size": null,
        "filename": "en-US",
        "parent": "C:\\Program Files\\Internet Explorer"
      },
      {
        "date": "12/07/2019",
        "time": "02:09 AM",
        "dir": false,
        "size": "54,784",
        "filename": "ExtExport.exe",
        "parent": "C:\\Program Files\\Internet Explorer"
      },
      ...
    ]
"""
import re
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.7'
    description = '`dir` command parser'
    author = 'Rasheed Elsaleh'
    author_email = 'rasheed@rebelliondefense.com'
    compatible = ['win32']
    tags = ['command']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary of Lists) raw structured data to process

    Returns:

        List of Dictionaries. Structured data to conform to the schema.
    """
    int_list = {'size'}

    for entry in proc_data:
        # add timestamps
        if 'date' in entry and 'time' in entry:
            dt = entry['date'] + ' ' + entry['time']
            timestamp = jc.utils.timestamp(dt, format_hint=(1600,))
            entry['epoch'] = timestamp.naive

        # add ints
        for key in entry:
            if key in int_list:
                entry[key] = jc.utils.convert_to_int(entry[key])

    return proc_data


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = []

    if jc.utils.has_data(data):

        for line in data.splitlines():
            if line.startswith(" Directory of "):
                parent_dir = line[len(" Directory of "):]
                continue
            # skip lines that don't start with a date
            if not re.match(r'^\d{2}/\d{2}/\d{4}', line):
                continue

            output_line = {}
            parsed_line = line.split()
            output_line["date"] = parsed_line[0]
            output_line["time"] = " ".join(parsed_line[1:3])
            output_line.setdefault("dir", False)
            output_line.setdefault("size", None)
            if parsed_line[3] == "<DIR>":
                output_line["dir"] = True
            else:
                output_line["size"] = parsed_line[3]

            output_line["filename"] = " ".join(parsed_line[4:])
            output_line["parent"] = parent_dir

            raw_output.append(output_line)

    if raw:
        return raw_output
    else:
        return _process(raw_output)



import os
import unittest
import json
import jc.parsers.x509_csr

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/x509-csr.der'), 'rb') as f:
        x509_csr_der = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/x509-csr.pem'), 'r', encoding='utf-8') as f:
        x509_csr_pem = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/x509-csr-windows.pem'), 'r', encoding='utf-8') as f:
        x509_csr_windows_pem = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/x509-csr-der.json'), 'r', encoding='utf-8') as f:
        x509_csr_der_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/x509-csr.json'), 'r', encoding='utf-8') as f:
        x509_csr_pem_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/x509-csr-windows.json'), 'r', encoding='utf-8') as f:
        x509_csr_windows_pem_json = json.loads(f.read())


    def test_x509_csr_nodata(self):
        """
        Test 'x509_csr' with no data
        """
        self.assertEqual(jc.parsers.x509_csr.parse('', quiet=True), [])

    def test_x509_csr_der(self):
        """
        Test csr file in DER format
        """
        self.assertEqual(jc.parsers.x509_csr.parse(self.x509_csr_der, quiet=True), self.x509_csr_der_json)

    def test_x509_csr_pem(self):
        """
        Test csr file in PEM format
        """
        self.assertEqual(jc.parsers.x509_csr.parse(self.x509_csr_pem, quiet=True), self.x509_csr_pem_json)

    def test_x509_csr_windows(self):
        """
        Test Windows csr file in PEM format
        """
        self.maxDiff = None
        self.assertEqual(jc.parsers.x509_csr.parse(self.x509_csr_windows_pem, quiet=True), self.x509_csr_windows_pem_json)


if __name__ == '__main__':
    unittest.main()



r"""jc - JSON Convert `ipconfig` Windows command output parser

Usage (cli):

    $ ipconfig /all | jc --ipconfig
    $ ipconfig | jc --ipconfig
    $ jc ipconfig /all

Usage (module):

    import jc
    result = jc.parse('ipconfig', ipconfig_command_output)

Schema:

    {
      "host_name":                             string,
      "primary_dns_suffix":                    string,
      "node_type":                             string,
      "ip_routing_enabled":                    boolean,
      "wins_proxy_enabled":                    boolean,
      "dns_suffix_search_list": [
                                               string
      ],
      "adapters": [
        {
          "name_long":                         string,
          "name":                              string,
          "type":                              string,
          "connection_specific_dns_suffix":    string,
          "connection_specific_dns_suffix_search_list": [
                                               string
          ]
          "description":                       string,
          "physical_address":                  string,
          "dhcp_enabled":                      boolean,
          "autoconfiguration_enabled":         boolean,
          "ipv6_addresses": [
            {
              "address":                       string,
              "status":                        string,
            },
          ],
          "temporary_ipv6_addresses": [
            {
              "address":                       string,
              "status":                        string,
            },
          ],
          "link_local_ipv6_addresses": [
            {
              "address":                       string,
              "status":                        string,
              "prefix_length":                 integer,
            }
          ],
          "ipv4_addresses": [
            {
              "address":                       string,     # [2]
              "subnet_mask":                   string,
              "status":                        string,
              "autoconfigured":                boolean     # [1]
            }
          ],
          "default_gateways": [
                                               string
          ],
          "dhcp_server": null,
          "dhcpv6_iaid":                       string,
          "dhcpv6_client_duid":                string,
          "dns_servers": [
                                               string
          ],
          "primary_wins_server":               string,
          "lease_expires":                     string,
          "lease_expires_epoch":               integer,    # [0]
          "lease_expires_iso":                 string,
          "lease_obtained":                    string,
          "lease_obtained_epoch":              integer,    # [0]
          "lease_obtained_iso":                string,
          "netbios_over_tcpip":                boolean,
          "media_state":                       string,
          "extras": [
            <string>:                          string
          ]
        }
      ],
      "extras": []
    }

    Notes:
      [0] - The epoch calculated timestamp field is naive. (i.e. based on
            the local time of the system the parser is run on)
      [1] - 'autoconfigured' under 'ipv4_address' is only providing
            indication if the ipv4 address was labeled as "Autoconfiguration
            IPv4 Address" vs "IPv4 Address". It does not infer any
            information from other fields
      [2] - Windows XP uses 'IP Address' instead of 'IPv4 Address'. Both
            values are parsed to the 'ipv4_address' object for consistency

Examples:

    $ ipconfig /all | jc --ipconfig -p
    {
      "host_name": "DESKTOP-WIN11-HOME",
      "primary_dns_suffix": null,
      "node_type": "Hybrid",
      "ip_routing_enabled": false,
      "wins_proxy_enabled": false,
      "dns_suffix_search_list": [
        "localdomain"
      ],
      "adapters": [
        {
          "name_long": "Ethernet adapter Ethernet",
          "name": "Ethernet",
          "type": "Ethernet",
          "connection_specific_dns_suffix": null,
          "connection_specific_dns_suffix_search_list": [],
          "description": "Intel(R) I211 Gigabit Network Connection",
          "physical_address": "24-4B-FE-AB-43-C3",
          "dhcp_enabled": true,
          "autoconfiguration_enabled": true,
          "ipv6_addresses": [],
          "temporary_ipv6_addresses": [],
          "link_local_ipv6_addresses": [],
          "ipv4_addresses": [],
          "default_gateways": [],
          "dhcp_server": null,
          "dhcpv6_iaid": null,
          "dhcpv6_client_duid": null,
          "dns_servers": [],
          "primary_wins_server": null,
          "lease_expires": null,
          "lease_obtained": null,
          "netbios_over_tcpip": null,
          "media_state": "Media disconnected",
          "extras": []
        },
        {
          "name_long": "Ethernet adapter Ethernet 2",
          "name": "Ethernet 2",
          "type": "Ethernet",
          "connection_specific_dns_suffix": null,
          "connection_specific_dns_suffix_search_list": [],
          "description": "Realtek PCIe 2.5GbE Family Controller",
          "physical_address": "24-4B-FE-57-3D-F2",
          "dhcp_enabled": true,
          "autoconfiguration_enabled": true,
          "ipv6_addresses": [],
          "temporary_ipv6_addresses": [],
          "link_local_ipv6_addresses": [],
          "ipv4_addresses": [],
          "default_gateways": [],
          "dhcp_server": null,
          "dhcpv6_iaid": null,
          "dhcpv6_client_duid": null,
          "dns_servers": [],
          "primary_wins_server": null,
          "lease_expires": null,
          "lease_obtained": null,
          "netbios_over_tcpip": null,
          "media_state": "Media disconnected",
          "extras": []
        },
        {
          "name_long": "Unknown adapter OpenVPN Data Channel Offload for NordVPN",
          "name": "OpenVPN Data Channel Offload for NordVPN",
          "type": "Unknown",
          "connection_specific_dns_suffix": null,
          "connection_specific_dns_suffix_search_list": [],
          "description": "OpenVPN Data Channel Offload",
          "physical_address": null,
          "dhcp_enabled": true,
          "autoconfiguration_enabled": true,
          "ipv6_addresses": [],
          "temporary_ipv6_addresses": [],
          "link_local_ipv6_addresses": [],
          "ipv4_addresses": [],
          "default_gateways": [],
          "dhcp_server": null,
          "dhcpv6_iaid": null,
          "dhcpv6_client_duid": null,
          "dns_servers": [],
          "primary_wins_server": null,
          "lease_expires": null,
          "lease_obtained": null,
          "netbios_over_tcpip": null,
          "media_state": "Media disconnected",
          "extras": []
        },
        {
          "name_long": "Unknown adapter Local Area Connection",
          "name": "Local Area Connection",
          "type": "Unknown",
          "connection_specific_dns_suffix": null,
          "connection_specific_dns_suffix_search_list": [],
          "description": "TAP-NordVPN Windows Adapter V9",
          "physical_address": "00-FF-4C-F4-5E-49",
          "dhcp_enabled": true,
          "autoconfiguration_enabled": true,
          "ipv6_addresses": [],
          "temporary_ipv6_addresses": [],
          "link_local_ipv6_addresses": [],
          "ipv4_addresses": [],
          "default_gateways": [],
          "dhcp_server": null,
          "dhcpv6_iaid": null,
          "dhcpv6_client_duid": null,
          "dns_servers": [],
          "primary_wins_server": null,
          "lease_expires": null,
          "lease_obtained": null,
          "netbios_over_tcpip": null,
          "media_state": "Media disconnected",
          "extras": []
        },
        {
          "name_long": "Wireless LAN adapter Local Area Connection* 1",
          "name": "Local Area Connection* 1",
          "type": "Wireless LAN",
          "connection_specific_dns_suffix": null,
          "connection_specific_dns_suffix_search_list": [],
          "description": "Microsoft Wi-Fi Direct Virtual Adapter",
          "physical_address": "A8-7E-EA-5A-7F-DE",
          "dhcp_enabled": true,
          "autoconfiguration_enabled": true,
          "ipv6_addresses": [],
          "temporary_ipv6_addresses": [],
          "link_local_ipv6_addresses": [],
          "ipv4_addresses": [],
          "default_gateways": [],
          "dhcp_server": null,
          "dhcpv6_iaid": null,
          "dhcpv6_client_duid": null,
          "dns_servers": [],
          "primary_wins_server": null,
          "lease_expires": null,
          "lease_obtained": null,
          "netbios_over_tcpip": null,
          "media_state": "Media disconnected",
          "extras": []
        },
        {
          "name_long": "Wireless LAN adapter Local Area Connection* 2",
          "name": "Local Area Connection* 2",
          "type": "Wireless LAN",
          "connection_specific_dns_suffix": null,
          "connection_specific_dns_suffix_search_list": [],
          "description": "Microsoft Wi-Fi Direct Virtual Adapter #2",
          "physical_address": "AA-7E-EA-F3-64-C3",
          "dhcp_enabled": true,
          "autoconfiguration_enabled": true,
          "ipv6_addresses": [],
          "temporary_ipv6_addresses": [],
          "link_local_ipv6_addresses": [],
          "ipv4_addresses": [],
          "default_gateways": [],
          "dhcp_server": null,
          "dhcpv6_iaid": null,
          "dhcpv6_client_duid": null,
          "dns_servers": [],
          "primary_wins_server": null,
          "lease_expires": null,
          "lease_obtained": null,
          "netbios_over_tcpip": null,
          "media_state": "Media disconnected",
          "extras": []
        },
        {
          "name_long": "Ethernet adapter VMware Network Adapter VMnet1",
          "name": "VMware Network Adapter VMnet1",
          "type": "Ethernet",
          "connection_specific_dns_suffix": null,
          "connection_specific_dns_suffix_search_list": [],
          "description": "VMware Virtual Ethernet Adapter for VMnet1",
          "physical_address": "00-50-56-CC-27-73",
          "dhcp_enabled": true,
          "autoconfiguration_enabled": true,
          "ipv6_addresses": [],
          "temporary_ipv6_addresses": [],
          "link_local_ipv6_addresses": [
            {
              "address": "fe80::f47d:9c7f:69dc:591e",
              "prefix_length": 8,
              "status": "Preferred"
            }
          ],
          "ipv4_addresses": [
            {
              "address": "192.168.181.1",
              "subnet_mask": "255.255.255.0",
              "status": "Preferred",
              "autoconfigured": false
            }
          ],
          "default_gateways": [],
          "dhcp_server": "192.168.181.254",
          "dhcpv6_iaid": "771772502",
          "dhcpv6_client_duid": "00-01-00-01-2C-CF-19-EB-24-4B-FE-5B-9B-E6",
          "dns_servers": [],
          "primary_wins_server": null,
          "lease_expires": "2024-09-19T18:01:29",
          "lease_obtained": "2024-09-19T08:31:29",
          "netbios_over_tcpip": true,
          "media_state": null,
          "extras": []
        },
        {
          "name_long": "Ethernet adapter VMware Network Adapter VMnet8",
          "name": "VMware Network Adapter VMnet8",
          "type": "Ethernet",
          "connection_specific_dns_suffix": null,
          "connection_specific_dns_suffix_search_list": [],
          "description": "VMware Virtual Ethernet Adapter for VMnet8",
          "physical_address": "00-50-56-C9-A3-78",
          "dhcp_enabled": true,
          "autoconfiguration_enabled": true,
          "ipv6_addresses": [],
          "temporary_ipv6_addresses": [],
          "link_local_ipv6_addresses": [
            {
              "address": "fe80::4551:bf0d:59dd:a4f0",
              "prefix_length": 10,
              "status": "Preferred"
            }
          ],
          "ipv4_addresses": [
            {
              "address": "192.168.213.1",
              "subnet_mask": "255.255.255.0",
              "status": "Preferred",
              "autoconfigured": false
            }
          ],
          "default_gateways": [],
          "dhcp_server": "192.168.213.254",
          "dhcpv6_iaid": "788549718",
          "dhcpv6_client_duid": "00-01-00-01-2C-CF-19-EB-24-4B-FE-5B-9B-E6",
          "dns_servers": [],
          "primary_wins_server": "192.168.213.2",
          "lease_expires": "2024-09-19T18:01:29",
          "lease_obtained": "2024-09-19T08:31:29",
          "netbios_over_tcpip": true,
          "media_state": null,
          "extras": []
        },
        {
          "name_long": "Wireless LAN adapter Wi-Fi",
          "name": "Wi-Fi",
          "type": "Wireless LAN",
          "connection_specific_dns_suffix": "localdomain",
          "connection_specific_dns_suffix_search_list": [],
          "description": "Intel(R) Wi-Fi 6 AX200 160MHz",
          "physical_address": "A8-7E-EA-55-26-B0",
          "dhcp_enabled": true,
          "autoconfiguration_enabled": true,
          "ipv6_addresses": [
            {
              "address": "fd63:cc9c:65eb:3f95:57c2:aa:10d8:db08",
              "status": "Preferred"
            }
          ],
          "temporary_ipv6_addresses": [
            {
              "address": "fd63:cc9c:65eb:3f95:8928:348e:d692:b7ef",
              "status": "Preferred"
            }
          ],
          "link_local_ipv6_addresses": [
            {
              "address": "fe80::4fae:1380:5a1b:8b6b",
              "prefix_length": 11,
              "status": "Preferred"
            }
          ],
          "ipv4_addresses": [
            {
              "address": "192.168.1.169",
              "subnet_mask": "255.255.255.0",
              "status": "Preferred",
              "autoconfigured": false
            }
          ],
          "default_gateways": [
            "192.168.1.1"
          ],
          "dhcp_server": "192.168.1.1",
          "dhcpv6_iaid": "162037482",
          "dhcpv6_client_duid": "00-01-00-01-2C-CF-19-EB-24-4B-FE-5B-9B-E6",
          "dns_servers": [
            "192.168.1.1"
          ],
          "primary_wins_server": null,
          "lease_expires": "2024-09-20T08:31:30",
          "lease_obtained": "2024-09-19T08:31:30",
          "netbios_over_tcpip": true,
          "media_state": null,
          "extras": []
        },
        {
          "name_long": "Ethernet adapter Bluetooth Network Connection",
          "name": "Bluetooth Network Connection",
          "type": "Ethernet",
          "connection_specific_dns_suffix": null,
          "connection_specific_dns_suffix_search_list": [],
          "description": "Bluetooth Device (Personal Area Network)",
          "physical_address": "A8-7E-EA-43-23-14",
          "dhcp_enabled": true,
          "autoconfiguration_enabled": true,
          "ipv6_addresses": [],
          "temporary_ipv6_addresses": [],
          "link_local_ipv6_addresses": [],
          "ipv4_addresses": [],
          "default_gateways": [],
          "dhcp_server": null,
          "dhcpv6_iaid": null,
          "dhcpv6_client_duid": null,
          "dns_servers": [],
          "primary_wins_server": null,
          "lease_expires": null,
          "lease_obtained": null,
          "netbios_over_tcpip": null,
          "media_state": "Media disconnected",
          "extras": []
        }
      ],
      "extras": []
    }
"""
from datetime import datetime
import re
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`ipconfig` Windows command parser'
    author = 'joehacksalot'
    author_email = 'joehacksalot@gmail.com'
    compatible = ['windows']
    magic_commands = ['ipconfig']
    tags = ['command']


__version__ = info.version


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Parsed dictionary. The raw and processed data structures are the same.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = {}
    if jc.utils.has_data(data):
        # Initialize the parsed output dictionary with all fields set to None or empty lists
        raw_output = _parse(data)

    return raw_output if raw else _process(raw_output)


def _process_ipv6_address(ip_address):
    address_split = ip_address["address"].split('%')
    try:
        if len(address_split) > 1:
            address = address_split[0]
            prefix_length = int(address_split[1])
        else:
            address = ip_address["address"]
            prefix_length = None
    except:
        address = ip_address["address"]
        prefix_length = None
    return {
              "address": address,
              "prefix_length": prefix_length,
              "status": ip_address["status"]
           }


def _process_ipv4_address(ip_address):
    autoconfigured = True if ip_address.get("autoconfigured","") is not None and 'autoconfigured' in ip_address.get("autoconfigured","") else False
    subnet_mask = ip_address["subnet_mask"]
    return {
              "address": ip_address["address"],
              "subnet_mask": subnet_mask,
              "status": ip_address["status"],
              "autoconfigured": autoconfigured
          }


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary) raw structured data to process

    Returns:

        Processed Dictionary. Structured data to conform to the schema.
    """
    processed = proc_data

    if "ip_routing_enabled" in processed and processed["ip_routing_enabled"] is not None:
        processed["ip_routing_enabled"] = (processed["ip_routing_enabled"].lower() == "yes")

    if "wins_proxy_enabled" in processed and processed["wins_proxy_enabled"] is not None:
        processed["wins_proxy_enabled"] = (processed["wins_proxy_enabled"].lower() == "yes")

    for adapter in processed["adapters"]:
        if "dhcp_enabled" in adapter and adapter["dhcp_enabled"] is not None:
            adapter["dhcp_enabled"] = (adapter["dhcp_enabled"].lower() == "yes")

        if "autoconfiguration_enabled" in adapter and adapter["autoconfiguration_enabled"] is not None:
            adapter["autoconfiguration_enabled"] = (adapter["autoconfiguration_enabled"].lower() == "yes")

        if "netbios_over_tcpip" in adapter and adapter["netbios_over_tcpip"] is not None:
            adapter["netbios_over_tcpip"] = (adapter["netbios_over_tcpip"].lower() == "enabled")

        if "lease_expires" in adapter and adapter["lease_expires"]:
            ts = jc.utils.timestamp(adapter['lease_expires'], format_hint=(1720,))
            adapter["lease_expires_epoch"] = ts.naive
            adapter["lease_expires_iso"] = ts.iso

        if "lease_obtained" in adapter and adapter["lease_obtained"]:
            ts = jc.utils.timestamp(adapter['lease_obtained'], format_hint=(1720,))
            adapter["lease_obtained_epoch"] = ts.naive
            adapter["lease_obtained_iso"] = ts.iso

        adapter["link_local_ipv6_addresses"] = [_process_ipv6_address(address) for address in adapter.get("link_local_ipv6_addresses", [])]
        adapter["ipv4_addresses"] = [_process_ipv4_address(address) for address in adapter.get("ipv4_addresses", [])]

    return processed


class _PushbackIterator:
    def __init__(self, iterator):
        self.iterator = iterator
        self.pushback_stack = []

    def __iter__(self):
        return self

    def __next__(self):
        if self.pushback_stack:
            return self.pushback_stack.pop()
        else:
            return next(self.iterator)

    def pushback(self, value):
        self.pushback_stack.append(value)


def _parse(data):
    # Initialize the parsed output dictionary with all fields set to None or empty lists
    parse_output = {
        "host_name": None,
        "primary_dns_suffix": None,
        "node_type": None,
        "ip_routing_enabled": None,
        "wins_proxy_enabled": None,
        "dns_suffix_search_list": [],
        "adapters": [],
        "extras": []  # To store unrecognized fields
    }

    lines = data.splitlines()
    lines = [line.rstrip() for line in lines if line.strip() != ""]

    line_iter = _PushbackIterator(iter(lines))
    adapter = None
    in_adapter_section = False

    for line in line_iter:
        line = line.rstrip()

        # Skip empty lines
        if not line.strip():
            continue

        # Header Section
        if not in_adapter_section:
            if "Windows IP Configuration" in line:
                continue
            elif _is_adapter_start_line(line):
                # Start of Adapter Section
                in_adapter_section = True
                adapter_name = line.strip(":").strip()
                adapter = _initialize_adapter(adapter_name)
                parse_output["adapters"].append(adapter)
            elif line.startswith("   "):
                key, value = _parse_line(line)
                if key:
                    _parse_header_line(parse_output, key, value, line_iter)
            else:
                continue
        else:
            # Adapter Sections
            if _is_adapter_start_line(line):
                # Start of new adapter
                adapter_name = line.strip(":").strip()
                adapter = _initialize_adapter(adapter_name)
                parse_output["adapters"].append(adapter)
            elif line.startswith("   "):
                key, value = _parse_line(line)
                if key:
                    _parse_adapter_line(adapter, key, value, line_iter)
            else:
                continue

    return parse_output


def _is_adapter_start_line(line):
    # Detect adapter start lines, e.g., "Ethernet adapter Ethernet:"
    return re.match(r"^[^\s].*adapter.*:", line, re.IGNORECASE)


def _initialize_adapter(adapter_name):
    adapter_name_split = adapter_name.split(" adapter ", 1)
    if len(adapter_name_split) > 1:
        adapter_type = adapter_name_split[0]
        adapter_short_name = adapter_name_split[1]
    else:
        adapter_type = None
        adapter_short_name = adapter_name

    # Initialize adapter dictionary with all fields set to None or empty lists
    return {
        "name_long": adapter_name,
        "name": adapter_short_name,
        "type": adapter_type,
        "connection_specific_dns_suffix": None,
        "connection_specific_dns_suffix_search_list": [],
        "description": None,
        "physical_address": None,
        "dhcp_enabled": None,
        "autoconfiguration_enabled": None,
        "ipv6_addresses": [],
        "temporary_ipv6_addresses": [],
        "link_local_ipv6_addresses": [],
        "ipv4_addresses": [],
        "default_gateways": [],
        "dhcp_server": None,
        "dhcpv6_iaid": None,
        "dhcpv6_client_duid": None,
        "dns_servers": [],
        "primary_wins_server": None,
        "lease_expires": None,
        "lease_obtained": None,
        "netbios_over_tcpip": None,
        "media_state": None,
        "extras": []  # To store unrecognized fields
    }


def _parse_line(line):
    # Split the line into key and value using ':' or multiple spaces
    key_value = re.split(r":", line.strip(), 1)
    if len(key_value) == 2:
        key, value = key_value
        key = key.strip().rstrip('. ')
        key = re.sub(r'[^\w]+', '_', key.lower())
        value = value.strip() if value.strip() != "" else None
        return key, value
    else:
        return None, None


def _parse_header_line(result, key, value, line_iter):
    if key in ["host_name", "primary_dns_suffix", "node_type", "ip_routing_enabled", "wins_proxy_enabled"]:
        result[key] = value
    elif key == "dns_suffix_search_list":
        if value:
            result["dns_suffix_search_list"].append(value)
        # Process additional entries
        _parse_additional_entries(result["dns_suffix_search_list"], line_iter)
    else:
        # Store unrecognized fields in extras
        result["extras"].append({key: value})


def _parse_adapter_line(adapter, key, value, line_iter):
    if key in ["connection_specific_dns_suffix","media_state", "description", "physical_address", "dhcp_enabled", 
               "autoconfiguration_enabled", "dhcpv6_iaid", "dhcpv6_client_duid", "netbios_over_tcpip", "dhcp_server", 
               "lease_obtained", "lease_expires", "primary_wins_server"]:
        adapter[key] = value

    elif key in ["ipv6_address", "temporary_ipv6_address", "link_local_ipv6_address"]:
        address_dict = _parse_ipv6_address(value)
        if key == "ipv6_address":
            adapter["ipv6_addresses"].append(address_dict)
        elif key == "temporary_ipv6_address":
            adapter["temporary_ipv6_addresses"].append(address_dict)
        elif key == "link_local_ipv6_address":
            adapter["link_local_ipv6_addresses"].append(address_dict)

    elif key in ["ipv4_address", "autoconfiguration_ipv4_address", "ip_address", "autoconfiguration_ip_address"]:
        ipv4_address_dict = _parse_ipv4_address(value, key, line_iter)
        adapter["ipv4_addresses"].append(ipv4_address_dict)

    elif key == "connection_specific_dns_suffix_search_list":
        if value:
            adapter["connection_specific_dns_suffix_search_list"].append(value)
        # Process additional connection specific dns suffix search list entries
        _parse_additional_entries(adapter["connection_specific_dns_suffix_search_list"], line_iter)

    elif key == "default_gateway":
        if value:
            adapter["default_gateways"].append(value)
        # Process additional gateways
        _parse_additional_entries(adapter["default_gateways"], line_iter)

    elif key == "dns_servers":
        if value:
            adapter["dns_servers"].append(value)
        # Process additional DNS servers
        _parse_additional_entries(adapter["dns_servers"], line_iter)

    elif key == "subnet_mask":
        # Subnet Mask should be associated with the last IPv4 address
        if adapter["ipv4_addresses"]:
            adapter["ipv4_addresses"][-1]["subnet_mask"] = value

    else:
        # Store unrecognized fields in extras
        adapter["extras"].append({key: value})


def _parse_ipv6_address(value):
    # Handle multiple status indicators
    match = re.match(r"([^\(]+)\((.*)\)", value) if value else None
    if match:
        address = match.group(1).strip()
        status = match.group(2).strip('()')
    else:
        address = value
        status = None
    return {
        "address": address,
        "status": status
    }


def _parse_ipv4_address(value, key, line_iter):
    # Handle autoconfigured status
    match = re.match(r"([^\(]+)\((.*)\)", value) if value else None
    if match:
        address = match.group(1).strip()
        status = match.group(2).strip('()')
        autoconfigured = 'autoconfigured' if 'autoconfiguration' in key or 'autoconfigured' in status else None
    else:
        address = value
        status = None
        autoconfigured = 'autoconfigured' if 'autoconfiguration' in key else None
    # Get subnet mask
    subnet_mask = None
    # Peek ahead for "Subnet Mask" line
    try:
        next_line = next(line_iter)
        next_key, next_value = _parse_line(next_line)
        if next_key == "subnet_mask":
            subnet_mask = next_value
        else:
            # If it's not "Subnet Mask", put it back into the iterator
            line_iter.pushback(next_line)
    except StopIteration:
        pass
    return {
            "address": address,
            "subnet_mask": subnet_mask,
            "autoconfigured": autoconfigured,
            "status": status
        }


def _parse_additional_entries(entry_list, line_iter):
    # Process additional lines that belong to the current entry (e.g., additional DNS servers, DNS Suffix Search List)
    while True:
        try:
            next_line = next(line_iter)
            if not next_line.strip():
                continue  # Skip empty lines

            # Check if the line is indented (starts with whitespace)
            if re.match(r"^\s\s\s\s", next_line):
                # It's an indented line; append the stripped line to entry_list
                entry_list.append(next_line.strip())
            else:
                # Not an indented line; push it back and exit
                line_iter.pushback(next_line)
                break
        except StopIteration:
            break



r"""jc - JSON Convert `needrestart -b` command output parser

Usage (cli):

    $ needrestart -b | jc --needrestart

or

    $ jc needrestart -b

Usage (module):

    import jc
    result = jc.parse('needrestart', needrestart_command_output)

Schema:

    {
      "version":                        string,
      "running_kernel_version":         string,
      "expected_kernel_version":        string,
      "kernel_status":                  integer,
      "container":                      string,
      "session": [
                                        string
      ],
      "service": [
                                        string
      ],
      "pid": [
                                        string
      ]
    }

Examples:

    $ needrestart -b | jc --needrestart -p
    {
      "version": "2.1",
      "running_kernel_version": "3.19.3-tl1+",
      "expected_kernel_version": "3.19.3-tl1+",
      "kernel_status": 1,
      "container": "LXC web1",
      "session": [
        "metabase @ user manager service",
        "root @ session #28017"
      ],
      "service": [
        "systemd-journald.service",
        "systemd-machined.service"
      ]
    }

    $ needrestart -b | jc --needrestart -p -r
    {
      "needrestart_ver": "2.1",
      "needrestart_kcur": "3.19.3-tl1+",
      "needrestart_kexp": "3.19.3-tl1+",
      "needrestart_ksta": "1",
      "needrestart_cont": "LXC web1",
      "needrestart_sess": [
        "metabase @ user manager service",
        "root @ session #28017"
      ],
      "needrestart_svc": [
        "systemd-journald.service",
        "systemd-machined.service"
      ]
    }
"""
from typing import List, Dict
from jc.jc_types import JSONDictType
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`needrestart -b` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    tags = ['command']
    magic_commands = ['needrestart -b']


__version__ = info.version


def _process(proc_data: JSONDictType) -> JSONDictType:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    key_map = {
        'needrestart_ver': 'version',
        'needrestart_kcur': 'running_kernel_version',
        'needrestart_kexp': 'expected_kernel_version',
        'needrestart_ksta': 'kernel_status',
        'needrestart_svc': 'service',
        'needrestart_cont': 'container',
        'needrestart_sess': 'session',
        'needrestart_pid': 'pid'
    }

    for key, val in proc_data.copy().items():
        if key == 'needrestart_ksta':
            proc_data[key] = jc.utils.convert_to_int(val)

        if key in key_map:
            proc_data[key_map[key]] = proc_data[key]
            del proc_data[key]

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> JSONDictType:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: Dict = {}
    sess_list: List[str] = []
    svc_list: List[str] = []
    pid_list: List[str] = []

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):

            if any([
                line.startswith('NEEDRESTART-VER'),
                line.startswith('NEEDRESTART-KCUR'),
                line.startswith('NEEDRESTART-KEXP'),
                line.startswith('NEEDRESTART-KSTA'),
                line.startswith('NEEDRESTART-CONT')
            ]):
                key, val = line.split(':', maxsplit=1)
                key = jc.utils.normalize_key(key)
                raw_output[key] = val.strip()
                continue

            if line.startswith('NEEDRESTART-SESS'):
                _, val = line.split(':', maxsplit=1)
                sess_list.append(val.strip())
                continue

            if line.startswith('NEEDRESTART-SVC'):
                _, val = line.split(':', maxsplit=1)
                svc_list.append(val.strip())
                continue

            if line.startswith('NEEDRESTART-PID'):
                _, val = line.split(':', maxsplit=1)
                pid_list.append(val.strip())
                continue

        if sess_list:
            raw_output['needrestart_sess'] = sess_list

        if svc_list:
            raw_output['needrestart_svc'] = svc_list

        if pid_list:
            raw_output['needrestart_pid'] = pid_list

    return raw_output if raw else _process(raw_output)



import os
import unittest
import json
import jc.parsers.kv

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/keyvalue.txt'), 'r', encoding='utf-8') as f:
        generic_ini_keyvalue = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/keyvalue-ifcfg.txt'), 'r', encoding='utf-8') as f:
        generic_ini_keyvalue_ifcfg = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/keyvalue.json'), 'r', encoding='utf-8') as f:
        generic_ini_keyvalue_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/keyvalue-ifcfg.json'), 'r', encoding='utf-8') as f:
        generic_ini_keyvalue_ifcfg_json = json.loads(f.read())


    def test_kv_nodata(self):
        """
        Test the test kv file with no data
        """
        self.assertEqual(jc.parsers.kv.parse('', quiet=True), {})

    def test_kv_keyvalue(self):
        """
        Test a file that only includes key/value lines
        """
        self.assertEqual(jc.parsers.kv.parse(self.generic_ini_keyvalue, quiet=True), self.generic_ini_keyvalue_json)

    def test_kv_keyvalue_ifcfg(self):
        """
        Test a sample ifcfg key/value file that has quotation marks in the values
        """
        self.assertEqual(jc.parsers.kv.parse(self.generic_ini_keyvalue_ifcfg, quiet=True), self.generic_ini_keyvalue_ifcfg_json)

    def test_kv_duplicate_keys(self):
        """
        Test input that contains duplicate keys. Only the last value should be used.
        """
        data = '''
duplicate_key: value1
another_key = foo
duplicate_key = value2
'''
        expected = {'duplicate_key': 'value2', 'another_key': 'foo'}
        self.assertEqual(jc.parsers.kv.parse(data, quiet=True), expected)

    def test_kv_doublequote(self):
        """
        Test kv string with double quotes around a value
        """
        data = '''
key1: "value1"
key2: value2
        '''
        expected = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual(jc.parsers.kv.parse(data, quiet=True), expected)

    def test_kv_singlequote(self):
        """
        Test kv string with double quotes around a value
        """
        data = '''
key1: 'value1'
key2: value2
        '''
        expected = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual(jc.parsers.kv.parse(data, quiet=True), expected)


if __name__ == '__main__':
    unittest.main()



r"""jc - JSON Convert `csv` implicit header file streaming parser

> This streaming parser outputs JSON Lines (cli) or returns an Iterable of
> Dictionaries (module)

The `csv` implicit header parser will attempt to automatically detect the
delimiter character. If the delimiter cannot be detected it will default to
comma. The file must have no header, and the field names will be generated
as "c0", "c1", etc.

> Note: The first 100 rows are read into memory to enable delimiter
> detection, then the rest of the rows are loaded lazily.

Usage (cli):

    $ cat file.csv | jc --csv-ih

Usage (module):

    import jc
    result = jc.parse('csv_ih', csv_output)

Schema:

CSV file converted to a Dictionary:
https://docs.python.org/3/library/csv.html

    {
      "c0":     string,
      "c1":     string,

      # below object only exists if using -qq or ignore_exceptions=True
      "_jc_meta": {
        "success":        boolean,     # false if error parsing
        "error":          string,      # exists if "success" is false
        "line":           string       # exists if "success" is false
      }
    }

Examples:

    $ cat homes.csv
    142, 160, 28, 10, 5, 3,  60, 0.28,  3167
    175, 180, 18,  8, 4, 1,  12, 0.43,  4033
    129, 132, 13,  6, 3, 1,  41, 0.33,  1471
    ...

    $ cat homes.csv | jc --csv-ih-s
    {"c0":"142","c1":"160","c2":"28","c3":"10","c4":"5"...}
    {"c0":"175","c1":"180","c2":"18","c3":"8","c4":"4"...}
    {"c0":"129","c1":"132","c2":"13","c3":"6","c4":"3"...}
    ...
"""
from typing import List, Union
from jc.jc_types import JSONDictType
import jc.parsers.csv_s as jc_csv_s
from jc.streaming import streaming_input_type_check
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = 'CSV implicit header file streaming parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    details = 'Using the python standard csv library'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    tags = ['standard', 'file', 'string']
    streaming = True


__version__ = info.version

def parse(
    data: Union[str, bytes],
    raw: bool = False,
    quiet: bool = False,
    ignore_exceptions: bool = False,
) -> List[JSONDictType]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    return jc_csv_s.parse(data, raw, quiet, ignore_exceptions, implicit_header=True)


r"""jc - JSON Convert `iftop` command output parser

Usage (cli):

    $ iftop -i <device> -t -B -s1 | jc --iftop

Usage (module):

    import jc
    result = jc.parse('iftop', iftop_command_output)

Schema:

    [
      {
        "device":                   string,
        "ip_address":               string,
        "mac_address":              string,
        "clients": [
          {
            "index":                integer,
            "connections": [
              {
                "host_name":        string,
                "host_port":        string,   # can be service or missing
                "last_2s":          integer,
                "last_10s":         integer,
                "last_40s":         integer,
                "cumulative":       integer,
                "direction":        string
              }
            ]
          }
        ]
        "total_send_rate": {
          "last_2s":                integer,
          "last_10s":               integer,
          "last_40s":               integer
        }
        "total_receive_rate": {
          "last_2s":                integer,
          "last_10s":               integer,
          "last_40s":               integer
        }
        "total_send_and_receive_rate": {
          "last_2s":                integer,
          "last_10s":               integer,
          "last_40s":               integer
        }
        "peak_rate": {
          "last_2s":                integer,
          "last_10s":               integer,
          "last_40s":               integer
        }
        "cumulative_rate": {
          "last_2s":                integer,
          "last_10s":               integer,
          "last_40s":               integer
        }
      }
    ]

Examples:

    $ iftop -i enp0s3 -t -P -s1 | jc --iftop -p
    [
      {
        "device": "enp0s3",
        "ip_address": "10.10.15.129",
        "mac_address": "08:00:27:c0:4a:4f",
        "clients": [
          {
            "index": 1,
            "connections": [
              {
                "host_name": "ubuntu-2004-clean-01",
                "host_port": "ssh",
                "last_2s": 448,
                "last_10s": 448,
                "last_40s": 448,
                "cumulative": 112,
                "direction": "send"
              },
              {
                "host_name": "10.10.15.72",
                "host_port": "40876",
                "last_2s": 208,
                "last_10s": 208,
                "last_40s": 208,
                "cumulative": 52,
                "direction": "receive"
              }
            ]
          }
        ],
        "total_send_rate": {
          "last_2s": 448,
          "last_10s": 448,
          "last_40s": 448
        },
        "total_receive_rate": {
          "last_2s": 208,
          "last_10s": 208,
          "last_40s": 208
        },
        "total_send_and_receive_rate": {
          "last_2s": 656,
          "last_10s": 656,
          "last_40s": 656
        },
        "peak_rate": {
          "last_2s": 448,
          "last_10s": 208,
          "last_40s": 656
        },
        "cumulative_rate": {
          "last_2s": 112,
          "last_10s": 52,
          "last_40s": 164
        }
      }
    ]

    $ iftop -i enp0s3 -t -P -s1 | jc --iftop -p -r
    [
      {
        "device": "enp0s3",
        "ip_address": "10.10.15.129",
        "mac_address": "11:22:33:44:55:66",
        "clients": [
          {
            "index": 1,
            "connections": [
              {
                "host_name": "ubuntu-2004-clean-01",
                "host_port": "ssh",
                "last_2s": "448b",
                "last_10s": "448b",
                "last_40s": "448b",
                "cumulative": "112B",
                "direction": "send"
              },
              {
                "host_name": "10.10.15.72",
                "host_port": "40876",
                "last_2s": "208b",
                "last_10s": "208b",
                "last_40s": "208b",
                "cumulative": "52B",
                "direction": "receive"
              }
            ]
          }
        ],
        "total_send_rate": {
          "last_2s": "448b",
          "last_10s": "448b",
          "last_40s": "448b"
        },
        "total_receive_rate": {
          "last_2s": "208b",
          "last_10s": "208b",
          "last_40s": "208b"
        },
        "total_send_and_receive_rate": {
          "last_2s": "656b",
          "last_10s": "656b",
          "last_40s": "656b"
        },
        "peak_rate": {
          "last_2s": "448b",
          "last_10s": "208b",
          "last_40s": "656b"
        },
        "cumulative_rate": {
          "last_2s": "112B",
          "last_10s": "52B",
          "last_40s": "164B"
        }
      }
    ]
"""
import re
from typing import List, Dict
from jc.jc_types import JSONDictType
import jc.utils


class info:
    """Provides parser metadata (version, author, etc.)"""
    version = "1.1"
    description = "`iftop` command parser"
    author = "Ron Green"
    author_email = "11993626+georgettica@users.noreply.github.com"
    compatible = ["linux"]
    tags = ["command"]


__version__ = info.version


def _process(proc_data: List[JSONDictType], quiet: bool = False) -> List[JSONDictType]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    string_to_bytes_fields = ["last_2s", "last_10s", "last_40s", "cumulative"]
    one_nesting = [
        "total_send_rate",
        "total_receive_rate",
        "total_send_and_receive_rate",
        "peak_rate",
        "cumulative_rate",
    ]

    if not proc_data:
        return proc_data
    for entry in proc_data:
        # print(f"{entry=}")
        for entry_key in entry:
            # print(f"{entry_key=}")
            if entry_key in one_nesting:
                # print(f"{entry[entry_key]=}")
                for one_nesting_item_key in entry[entry_key]:
                    # print(f"{one_nesting_item_key=}")
                    if one_nesting_item_key in string_to_bytes_fields:
                        entry[entry_key][one_nesting_item_key] = jc.utils.convert_size_to_int(entry[entry_key][one_nesting_item_key])
            elif entry_key == "clients":
                for client in entry[entry_key]:
                    # print(f"{client=}")
                    if "connections" not in client:
                        continue
                    for connection in client["connections"]:
                        # print(f"{connection=}")
                        for connection_key in connection:
                            # print(f"{connection_key=}")
                            if connection_key in string_to_bytes_fields:
                                connection[connection_key] = jc.utils.convert_size_to_int(connection[connection_key])
    return proc_data


def parse(data: str, raw: bool = False, quiet: bool = False) -> List[JSONDictType]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List[Dict] = []
    interface_item: Dict = {}
    current_client: Dict = {}
    clients: List = []
    is_previous_line_interface = False
    saw_already_host_line = False

    before_arrow = r"\s+(?P<index>\d+)\s+(?P<host_name>[^\s]+):(?P<host_port>[^\s]+)\s+"
    before_arrow_no_port = r"\s+(?P<index>\d+)\s+(?P<host_name>[^\s]+)\s+"
    after_arrow_before_newline = r"\s+(?P<send_last_2s>[^\s]+)\s+(?P<send_last_10s>[^\s]+)\s+(?P<send_last_40s>[^\s]+)\s+(?P<send_cumulative>[^\s]+)"
    newline_before_arrow = r"\s+(?P<receive_ip>.+):(?P<receive_port>\w+)\s+"
    newline_before_arrow_no_port = r"\s+(?P<receive_ip>.+)\s+"
    after_arrow_till_end = r"\s+(?P<receive_last_2s>[^\s]+)\s+(?P<receive_last_10s>[^\s]+)\s+(?P<receive_last_40s>[^\s]+)\s+(?P<receive_cumulative>[^\s]+)"
    re_linux_clients_before_newline = re.compile(
        rf"{before_arrow}=>{after_arrow_before_newline}"
    )
    re_linux_clients_before_newline_no_port = re.compile(
        rf"{before_arrow_no_port}=>{after_arrow_before_newline}"
    )
    re_linux_clients_after_newline_no_port = re.compile(
        rf"{newline_before_arrow_no_port}<={after_arrow_till_end}"
    )

    re_linux_clients_after_newline = re.compile(
        rf"{newline_before_arrow}<={after_arrow_till_end}"
    )

    re_total_send_rate = re.compile(
        r"Total send rate:\s+(?P<total_send_rate_last_2s>[^\s]+)\s+(?P<total_send_rate_last_10s>[^\s]+)\s+(?P<total_send_rate_last_40s>[^\s]+)"
    )
    re_total_receive_rate = re.compile(
        r"Total receive rate:\s+(?P<total_receive_rate_last_2s>[^\s]+)\s+(?P<total_receive_rate_last_10s>[^\s]+)\s+(?P<total_receive_rate_last_40s>[^\s]+)"
    )
    re_total_send_and_receive_rate = re.compile(
        r"Total send and receive rate:\s+(?P<total_send_and_receive_rate_last_2s>[^\s]+)\s+(?P<total_send_and_receive_rate_last_10s>[^\s]+)\s+(?P<total_send_and_receive_rate_last_40s>[^\s]+)"
    )
    re_peak_rate = re.compile(
        r"Peak rate \(sent/received/total\):\s+(?P<peak_rate_sent>[^\s]+)\s+(?P<peak_rate_received>[^\s]+)\s+(?P<peak_rate_total>[^\s]+)"
    )
    re_cumulative_rate = re.compile(
        r"Cumulative \(sent/received/total\):\s+(?P<cumulative_rate_sent>[^\s]+)\s+(?P<cumulative_rate_received>[^\s]+)\s+(?P<cumulative_rate_total>[^\s]+)"
    )

    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    if not jc.utils.has_data(data):
        return raw_output if raw else _process(raw_output, quiet=quiet)

    for line in filter(None, data.splitlines()):
        if line.startswith("interface:"):
            # Example:
            # interface: enp0s3
            interface_item["device"] = line.split(":")[1].strip()

        elif line.startswith("IP address is:"):
            # Example:
            # IP address is: 10.10.15.129
            interface_item["ip_address"] = line.split(":")[1].strip()

        elif line.startswith("MAC address is:"):
            # Example:
            # MAC address is: 08:00:27:c0:4a:4f
            # strip off the "MAC address is: " part
            data_without_front_list = line.split(":")[1:]

            # join the remaining parts back together
            data_without_front = ":".join(data_without_front_list)
            interface_item["mac_address"] = data_without_front.strip()

        elif line.startswith("Listening on"):
            # Example:
            # Listening on enp0s3
            pass

        elif (
            line.startswith("# Host name (port/service if enabled)")
            and not saw_already_host_line
        ):
            saw_already_host_line = True
            # Example:
            #    # Host name (port/service if enabled)            last 2s   last 10s   last 40s cumulative
            pass

        elif (
            line.startswith("# Host name (port/service if enabled)")
            and saw_already_host_line
        ):
            old_interface_item, interface_item = interface_item, {}
            interface_item.update(
                {
                    "device": old_interface_item["device"],
                    "ip_address": old_interface_item["ip_address"],
                    "mac_address": old_interface_item["mac_address"],
                }
            )

        elif "=>" in line and is_previous_line_interface and ":" in line:
            # should not happen
            pass

        elif "=>" in line and not is_previous_line_interface and ":" in line:
            # Example:
            #    1 ubuntu-2004-clean-01:ssh                 =>       448b       448b       448b       112B
            is_previous_line_interface = True
            match_raw = re_linux_clients_before_newline.match(line)

            if not match_raw:
                # this is a bug in iftop
                continue

            match_dict = match_raw.groupdict()
            current_client = {}
            current_client["index"] = int(match_dict["index"])
            current_client["connections"] = []
            current_client_send = {
                "host_name": match_dict["host_name"],
                "host_port": match_dict["host_port"],
                "last_2s": match_dict["send_last_2s"],
                "last_10s": match_dict["send_last_10s"],
                "last_40s": match_dict["send_last_40s"],
                "cumulative": match_dict["send_cumulative"],
                "direction": "send",
            }
            current_client["connections"].append(current_client_send)
            # not adding yet as the receive part is not yet parsed

        elif "=>" in line and not is_previous_line_interface and ":" not in line:
            # should not happen
            pass

        elif "=>" in line and is_previous_line_interface and ":" not in line:
            is_previous_line_interface = True
            match_raw = re_linux_clients_before_newline_no_port.match(line)

            if not match_raw:
                # this is a bug in iftop
                continue

            match_dict = match_raw.groupdict()
            current_client = {}
            current_client["index"] = int(match_dict["index"])
            current_client["connections"] = []
            current_client_send = {
                "host_name": match_dict["host_name"],
                "last_2s": match_dict["send_last_2s"],
                "last_10s": match_dict["send_last_10s"],
                "last_40s": match_dict["send_last_40s"],
                "cumulative": match_dict["send_cumulative"],
                "direction": "send",
            }
            current_client["connections"].append(current_client_send)
            # not adding yet as the receive part is not yet parsed

        elif "<=" in line and not is_previous_line_interface and ":" in line:
            # should not happen
            pass

        elif "<=" in line and is_previous_line_interface and ":" in line:
            # Example:
            #      10.10.15.72:40876                        <=       208b       208b       208b        52B
            is_previous_line_interface = False
            match_raw = re_linux_clients_after_newline.match(line)

            if not match_raw:
                # this is a bug in iftop
                continue

            match_dict = match_raw.groupdict()
            current_client_receive = {
                "host_name": match_dict["receive_ip"],
                "host_port": match_dict["receive_port"],
                "last_2s": match_dict["receive_last_2s"],
                "last_10s": match_dict["receive_last_10s"],
                "last_40s": match_dict["receive_last_40s"],
                "cumulative": match_dict["receive_cumulative"],
                "direction": "receive",
            }

            current_client["connections"].append(current_client_receive)
            clients.append(current_client)

        elif "<=" in line and not is_previous_line_interface and ":" not in line:
            # should not happen
            pass

        elif "<=" in line and is_previous_line_interface and ":" not in line:
            # Example:
            #      10.10.15.72:40876                        <=       208b       208b       208b        52B
            is_previous_line_interface = False
            match_raw = re_linux_clients_after_newline_no_port.match(line)

            if not match_raw:
                # this is a bug in iftop
                continue

            match_dict = match_raw.groupdict()
            current_client_receive = {
                "host_name": match_dict["receive_ip"],
                "last_2s": match_dict["receive_last_2s"],
                "last_10s": match_dict["receive_last_10s"],
                "last_40s": match_dict["receive_last_40s"],
                "cumulative": match_dict["receive_cumulative"],
                "direction": "receive",
            }

            current_client["connections"].append(current_client_receive)
            clients.append(current_client)

        # check if all of the characters are dashes or equal signs
        elif all(c == "-" for c in line):
            pass

        elif line.startswith("Total send rate"):
            # Example:
            # Total send rate:                                       448b       448b       448b
            match_raw = re_total_send_rate.match(line)

            if not match_raw:
                # this is a bug in iftop
                continue

            match_dict = match_raw.groupdict()
            interface_item["total_send_rate"] = {}
            interface_item["total_send_rate"].update(
                {
                    "last_2s": match_dict["total_send_rate_last_2s"],
                    "last_10s": match_dict["total_send_rate_last_10s"],
                    "last_40s": match_dict["total_send_rate_last_40s"],
                }
            )

        elif line.startswith("Total receive rate"):
            # Example:
            # Total receive rate:                                    208b       208b       208b
            match_raw = re_total_receive_rate.match(line)

            if not match_raw:
                # this is a bug in iftop
                continue

            match_dict = match_raw.groupdict()
            interface_item["total_receive_rate"] = {}
            interface_item["total_receive_rate"].update(
                {
                    "last_2s": match_dict["total_receive_rate_last_2s"],
                    "last_10s": match_dict["total_receive_rate_last_10s"],
                    "last_40s": match_dict["total_receive_rate_last_40s"],
                }
            )

        elif line.startswith("Total send and receive rate"):
            # Example:
            # Total send and receive rate:                           656b       656b       656b
            match_raw = re_total_send_and_receive_rate.match(line)

            if not match_raw:
                # this is a bug in iftop
                continue

            match_dict = match_raw.groupdict()
            interface_item["total_send_and_receive_rate"] = {}
            interface_item["total_send_and_receive_rate"].update(
                {
                    "last_2s": match_dict["total_send_and_receive_rate_last_2s"],
                    "last_10s": match_dict["total_send_and_receive_rate_last_10s"],
                    "last_40s": match_dict["total_send_and_receive_rate_last_40s"],
                }
            )

        elif line.startswith("Peak rate"):
            match_raw = re_peak_rate.match(line)

            if not match_raw:
                # this is a bug in iftop
                continue

            match_dict = match_raw.groupdict()
            interface_item["peak_rate"] = {}
            interface_item["peak_rate"].update(
                {
                    "last_2s": match_dict["peak_rate_sent"],
                    "last_10s": match_dict["peak_rate_received"],
                    "last_40s": match_dict["peak_rate_total"],
                }
            )

        elif line.startswith("Cumulative"):
            match_raw = re_cumulative_rate.match(line)

            if not match_raw:
                # this is a bug in iftop
                continue

            match_dict = match_raw.groupdict()
            interface_item["cumulative_rate"] = {}
            interface_item["cumulative_rate"].update(
                {
                    "last_2s": match_dict["cumulative_rate_sent"],
                    "last_10s": match_dict["cumulative_rate_received"],
                    "last_40s": match_dict["cumulative_rate_total"],
                }
            )

        elif all(c == "=" for c in line):
            interface_item["clients"] = clients
            clients = []
            # keep the copy here as without it keeps the objects linked
            raw_output.append(interface_item.copy())

    return raw_output if raw else _process(raw_output, quiet=quiet)



r"""jc - JSON Convert `zpool status` command output parser

Works with or without the `-v` option.

Usage (cli):

    $ zpool status | jc --zpool-status

or

    $ jc zpool status

Usage (module):

    import jc
    result = jc.parse('zpool_status', zpool_status_command_output)

Schema:

    [
      {
        "pool":                               string,
        "state":                              string,
        "status":                             string,
        "action":                             string,
        "see":                                string,
        "scan":                               string,
        "scrub":                              string,
        "config": [
          {
            "name":                           string,
            "state":                          string/null,
            "read":                           integer/null,
            "write":                          integer/null,
            "checksum":                       integer/null,
            "errors":                         string/null,
          }
        ],
        "errors":                             string
      }
    ]

Examples:

    $ zpool status -v | jc --zpool-status -p
    [
      {
        "pool": "tank",
        "state": "DEGRADED",
        "status": "One or more devices could not be opened.  Suffic...",
        "action": "Attach the missing device and online it using 'zpool...",
        "see": "http://www.sun.com/msg/ZFS-8000-2Q",
        "scrub": "none requested",
        "config": [
          {
            "name": "tank",
            "state": "DEGRADED",
            "read": 0,
            "write": 0,
            "checksum": 0
          },
          {
            "name": "mirror-0",
            "state": "DEGRADED",
            "read": 0,
            "write": 0,
            "checksum": 0
          },
          {
            "name": "c1t0d0",
            "state": "ONLINE",
            "read": 0,
            "write": 0,
            "checksum": 0
          },
          {
            "name": "c1t1d0",
            "state": "UNAVAIL",
            "read": 0,
            "write": 0,
            "checksum": 0,
            "errors": "cannot open"
          }
        ],
        "errors": "No known data errors"
      }
    ]

    $ zpool status -v | jc --zpool-status -p -r
    [
      {
        "pool": "tank",
        "state": "DEGRADED",
        "status": "One or more devices could not be opened.  Sufficient...",
        "action": "Attach the missing device and online it using 'zpool...",
        "see": "http://www.sun.com/msg/ZFS-8000-2Q",
        "scrub": "none requested",
        "config": [
          {
            "name": "tank",
            "state": "DEGRADED",
            "read": "0",
            "write": "0",
            "checksum": "0"
          },
          {
            "name": "mirror-0",
            "state": "DEGRADED",
            "read": "0",
            "write": "0",
            "checksum": "0"
          },
          {
            "name": "c1t0d0",
            "state": "ONLINE",
            "read": "0",
            "write": "0",
            "checksum": "0"
          },
          {
            "name": "c1t1d0",
            "state": "UNAVAIL",
            "read": "0",
            "write": "0",
            "checksum": "0",
            "errors": "cannot open"
          }
        ],
        "errors": "No known data errors"
      }
    ]
"""
from typing import List, Dict
from jc.jc_types import JSONDictType
import jc.utils
from jc.parsers.kv import parse as kv_parse


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.2'
    description = '`zpool status` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'freebsd']
    tags = ['command']
    magic_commands = ['zpool status']


__version__ = info.version


def _process(proc_data: List[JSONDictType]) -> List[JSONDictType]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    int_list = {'read', 'write', 'checksum'}

    for obj in proc_data:
        if 'config' in obj:
            for conf in obj['config']:
                for k, v in conf.items():
                    if k in int_list:
                        conf[k] = jc.utils.convert_to_int(v)

    return proc_data


def _build_config_list(string: str) -> List[Dict]:
    config_list: List = []
    for line in filter(None, string.splitlines()):
        if line.strip().endswith('READ WRITE CKSUM'):
            continue

        line_list = line.strip().split(maxsplit=5)
        config_obj: Dict = {}
        config_obj['name'] = line_list[0] if len(line_list) > 0 else None
        config_obj['state'] = line_list[1] if len(line_list) > 1 else None
        config_obj['read'] = line_list[2] if len(line_list) > 2 else None
        config_obj['write'] = line_list[3] if len(line_list) > 3 else None
        config_obj['checksum'] = line_list[4] if len(line_list) > 4 else None
        if len(line_list) == 6:
            config_obj['errors'] = line_list[5]
        config_list.append(config_obj)

    return config_list

def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[JSONDictType]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List[Dict] = []
    pool_str: str = ''
    pool_obj: Dict = {}

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):

            if line.lstrip().startswith('pool: '):
                if pool_str:
                    pool_obj = kv_parse(pool_str)
                    if 'config' in pool_obj:
                        pool_obj['config'] = _build_config_list(pool_obj['config'])
                    raw_output.append(pool_obj)
                pool_str = ''
                pool_str += line + '\n'
                continue

            # preserve indentation in continuation lines
            if line.startswith('        ') or line.startswith('\t'):
                pool_str += line + '\n'
                continue

            # indent path lines for errors field
            if line.startswith('/'):
                pool_str += '  ' + line + '\n'
                continue

            # remove initial spaces from field start lines so we don't confuse line continuation
            pool_str += line.strip() + '\n'

    if pool_str:
        pool_obj = kv_parse(pool_str)
        if 'config' in pool_obj:
            pool_obj['config'] = _build_config_list(pool_obj['config'])
        raw_output.append(pool_obj)

    return raw_output if raw else _process(raw_output)



r"""jc - JSON Convert `/proc/loadavg` file parser

Usage (cli):

    $ cat /proc/loadavg | jc --proc

or

    $ jc /proc/loadavg

or

    $ cat /proc/loadavg | jc --proc-loadavg

Usage (module):

    import jc
    result = jc.parse('proc', proc_loadavg_file)

or

    import jc
    result = jc.parse('proc_loadavg', proc_loadavg_file)

Schema:

All values are integers.

    {
      "load_1m":              float,
      "load_5m":              float,
      "load_15m":             float,
      "running":              integer,
      "available":            integer,
      "last_pid":             integer
    }

Examples:

    $ cat /proc/loadavg | jc --proc -p
    {
      "load_1m": 0.0,
      "load_5m": 0.01,
      "load_15m": 0.03,
      "running": 2,
      "available": 111,
      "last_pid": 2039
    }

    $ cat /proc/loadavg | jc --proc -p -r
    {
      "load_1m": "0.00",
      "load_5m": "0.01",
      "load_15m": "0.03",
      "running": "2",
      "available": "111",
      "last_pid": "2039"
    }
"""
from typing import Dict
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`/proc/loadavg` file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    tags = ['file']
    hidden = True


__version__ = info.version


def _process(proc_data: Dict) -> Dict:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary) raw structured data to process

    Returns:

        Dictionary. Structured to conform to the schema.
    """
    float_list = {'load_1m', 'load_5m', 'load_15m'}
    int_list = {'running', 'available', 'last_pid'}

    for key in proc_data:
        if key in float_list:
            proc_data[key] = float(proc_data[key])

        if key in int_list:
            proc_data[key] = int(proc_data[key])

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> Dict:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: Dict = {}

    if jc.utils.has_data(data):

        load_1m, load_5m, load_15m, runnable, last_pid = data.split()
        running, available = runnable.split('/')

        raw_output = {
            'load_1m': load_1m,
            'load_5m': load_5m,
            'load_15m': load_15m,
            'running': running,
            'available': available,
            'last_pid': last_pid
        }

    return raw_output if raw else _process(raw_output)



import os
import json
import unittest
import jc.parsers.ss

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/ss-sudo-a.out'), 'r', encoding='utf-8') as f:
        centos_7_7_ss_sudo_a = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/ss-sudo-a.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_ss_sudo_a = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/ss-sudo-tulpen.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_ss_sudo_tulpen = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/ss-wide.out'), 'r', encoding='utf-8') as f:
        ss_wide = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/ss-sudo-a.json'), 'r', encoding='utf-8') as f:
        centos_7_7_ss_sudo_a_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/ss-sudo-a.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_ss_sudo_a_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/ss-sudo-tulpen.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_ss_sudo_tulpen_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/ss-wide.json'), 'r', encoding='utf-8') as f:
        ss_wide_json = json.loads(f.read())

    def test_ss_nodata(self):
        """
        Test 'ss' with no data
        """
        self.assertEqual(jc.parsers.ss.parse('', quiet=True), [])

    def test_ss_sudo_a_centos_7_7(self):
        """
        Test 'sudo ss -a' on Centos 7.7
        """
        self.assertEqual(jc.parsers.ss.parse(self.centos_7_7_ss_sudo_a, quiet=True), self.centos_7_7_ss_sudo_a_json)

    def test_ss_sudo_a_ubuntu_18_4(self):
        """
        Test 'sudo ss -a' on Ubuntu 18.4
        """
        self.assertEqual(jc.parsers.ss.parse(self.ubuntu_18_4_ss_sudo_a, quiet=True), self.ubuntu_18_4_ss_sudo_a_json)

    def test_ss_sudo_tulpen_ubuntu_18_4(self):
        """
        Test 'sudo ss -tulpen' on Ubuntu 18.4
        """
        self.assertEqual(jc.parsers.ss.parse(self.ubuntu_18_4_ss_sudo_tulpen, quiet=True), self.ubuntu_18_4_ss_sudo_tulpen_json)

    def test_ss_wide(self):
        """
        Test 'sudo ss' with wide format and lots of options
        """
        self.assertEqual(jc.parsers.ss.parse(self.ss_wide, quiet=True), self.ss_wide_json)

if __name__ == '__main__':
    unittest.main()



import os
import unittest
import json
import jc.parsers.airport

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/airport-I.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_airport_I = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/airport-I.json'), 'r', encoding='utf-8') as f:
        osx_10_14_6_airport_I_json = json.loads(f.read())

    def test_airport_I_nodata(self):
        """
        Test 'airport -I' with no data
        """
        self.assertEqual(jc.parsers.airport.parse('', quiet=True), {})

    def test_airport_I_osx_10_14_6(self):
        """
        Test 'airport -I' on OSX 10.14.6
        """
        self.assertEqual(jc.parsers.airport.parse(self.osx_10_14_6_airport_I, quiet=True), self.osx_10_14_6_airport_I_json)


if __name__ == '__main__':
    unittest.main()



# coding: utf-8

"""
Functions for parsing and dumping using the ASN.1 DER encoding. Exports the
following items:

 - emit()
 - parse()
 - peek()

Other type classes are defined that help compose the types listed above.
"""

from __future__ import unicode_literals, division, absolute_import, print_function

import sys

from ._types import chr_cls, type_name
from .util import int_from_bytes, int_to_bytes

_INSUFFICIENT_DATA_MESSAGE = 'Insufficient data - %s bytes requested but only %s available'
_MAX_DEPTH = 10


def emit(class_, method, tag, contents):
    """
    Constructs a byte string of an ASN.1 DER-encoded value

    This is typically not useful. Instead, use one of the standard classes from
    asn1crypto.core, or construct a new class with specific fields, and call the
    .dump() method.

    :param class_:
        An integer ASN.1 class value: 0 (universal), 1 (application),
        2 (context), 3 (private)

    :param method:
        An integer ASN.1 method value: 0 (primitive), 1 (constructed)

    :param tag:
        An integer ASN.1 tag value

    :param contents:
        A byte string of the encoded byte contents

    :return:
        A byte string of the ASN.1 DER value (header and contents)
    """

    if not isinstance(class_, int):
        raise TypeError('class_ must be an integer, not %s' % type_name(class_))

    if class_ < 0 or class_ > 3:
        raise ValueError('class_ must be one of 0, 1, 2 or 3, not %s' % class_)

    if not isinstance(method, int):
        raise TypeError('method must be an integer, not %s' % type_name(method))

    if method < 0 or method > 1:
        raise ValueError('method must be 0 or 1, not %s' % method)

    if not isinstance(tag, int):
        raise TypeError('tag must be an integer, not %s' % type_name(tag))

    if tag < 0:
        raise ValueError('tag must be greater than zero, not %s' % tag)

    if not isinstance(contents, bytes):
        raise TypeError('contents must be a byte string, not %s' % type_name(contents))

    return _dump_header(class_, method, tag, contents) + contents


def parse(contents, strict=False):
    """
    Parses a byte string of ASN.1 BER/DER-encoded data.

    This is typically not useful. Instead, use one of the standard classes from
    asn1crypto.core, or construct a new class with specific fields, and call the
    .load() class method.

    :param contents:
        A byte string of BER/DER-encoded data

    :param strict:
        A boolean indicating if trailing data should be forbidden - if so, a
        ValueError will be raised when trailing data exists

    :raises:
        ValueError - when the contents do not contain an ASN.1 header or are truncated in some way
        TypeError - when contents is not a byte string

    :return:
        A 6-element tuple:
         - 0: integer class (0 to 3)
         - 1: integer method
         - 2: integer tag
         - 3: byte string header
         - 4: byte string content
         - 5: byte string trailer
    """

    if not isinstance(contents, bytes):
        raise TypeError('contents must be a byte string, not %s' % type_name(contents))

    contents_len = len(contents)
    info, consumed = _parse(contents, contents_len)
    if strict and consumed != contents_len:
        raise ValueError('Extra data - %d bytes of trailing data were provided' % (contents_len - consumed))
    return info


def peek(contents):
    """
    Parses a byte string of ASN.1 BER/DER-encoded data to find the length

    This is typically used to look into an encoded value to see how long the
    next chunk of ASN.1-encoded data is. Primarily it is useful when a
    value is a concatenation of multiple values.

    :param contents:
        A byte string of BER/DER-encoded data

    :raises:
        ValueError - when the contents do not contain an ASN.1 header or are truncated in some way
        TypeError - when contents is not a byte string

    :return:
        An integer with the number of bytes occupied by the ASN.1 value
    """

    if not isinstance(contents, bytes):
        raise TypeError('contents must be a byte string, not %s' % type_name(contents))

    info, consumed = _parse(contents, len(contents))
    return consumed


def _parse(encoded_data, data_len, pointer=0, lengths_only=False, depth=0):
    """
    Parses a byte string into component parts

    :param encoded_data:
        A byte string that contains BER-encoded data

    :param data_len:
        The integer length of the encoded data

    :param pointer:
        The index in the byte string to parse from

    :param lengths_only:
        A boolean to cause the call to return a 2-element tuple of the integer
        number of bytes in the header and the integer number of bytes in the
        contents. Internal use only.

    :param depth:
        The recursion depth when evaluating indefinite-length encoding.

    :return:
        A 2-element tuple:
         - 0: A tuple of (class_, method, tag, header, content, trailer)
         - 1: An integer indicating how many bytes were consumed
    """

    if depth > _MAX_DEPTH:
        raise ValueError('Indefinite-length recursion limit exceeded')

    start = pointer

    if data_len < pointer + 1:
        raise ValueError(_INSUFFICIENT_DATA_MESSAGE % (1, data_len - pointer))
    first_octet = encoded_data[pointer]

    pointer += 1

    tag = first_octet & 31
    constructed = (first_octet >> 5) & 1
    # Base 128 length using 8th bit as continuation indicator
    if tag == 31:
        tag = 0
        while True:
            if data_len < pointer + 1:
                raise ValueError(_INSUFFICIENT_DATA_MESSAGE % (1, data_len - pointer))
            num = encoded_data[pointer]
            pointer += 1
            if num == 0x80 and tag == 0:
                raise ValueError('Non-minimal tag encoding')
            tag *= 128
            tag += num & 127
            if num >> 7 == 0:
                break
        if tag < 31:
            raise ValueError('Non-minimal tag encoding')

    if data_len < pointer + 1:
        raise ValueError(_INSUFFICIENT_DATA_MESSAGE % (1, data_len - pointer))
    length_octet = encoded_data[pointer]
    pointer += 1
    trailer = b''

    if length_octet >> 7 == 0:
        contents_end = pointer + (length_octet & 127)

    else:
        length_octets = length_octet & 127
        if length_octets:
            if data_len < pointer + length_octets:
                raise ValueError(_INSUFFICIENT_DATA_MESSAGE % (length_octets, data_len - pointer))
            pointer += length_octets
            contents_end = pointer + int_from_bytes(encoded_data[pointer - length_octets:pointer], signed=False)

        else:
            # To properly parse indefinite length values, we need to scan forward
            # parsing headers until we find a value with a length of zero. If we
            # just scanned looking for \x00\x00, nested indefinite length values
            # would not work.
            if not constructed:
                raise ValueError('Indefinite-length element must be constructed')
            contents_end = pointer
            while data_len < contents_end + 2 or encoded_data[contents_end:contents_end+2] != b'\x00\x00':
                _, contents_end = _parse(encoded_data, data_len, contents_end, lengths_only=True, depth=depth+1)
            contents_end += 2
            trailer = b'\x00\x00'

    if contents_end > data_len:
        raise ValueError(_INSUFFICIENT_DATA_MESSAGE % (contents_end - pointer, data_len - pointer))

    if lengths_only:
        return (pointer, contents_end)

    return (
        (
            first_octet >> 6,
            constructed,
            tag,
            encoded_data[start:pointer],
            encoded_data[pointer:contents_end-len(trailer)],
            trailer
        ),
        contents_end
    )


def _dump_header(class_, method, tag, contents):
    """
    Constructs the header bytes for an ASN.1 object

    :param class_:
        An integer ASN.1 class value: 0 (universal), 1 (application),
        2 (context), 3 (private)

    :param method:
        An integer ASN.1 method value: 0 (primitive), 1 (constructed)

    :param tag:
        An integer ASN.1 tag value

    :param contents:
        A byte string of the encoded byte contents

    :return:
        A byte string of the ASN.1 DER header
    """

    header = b''

    id_num = 0
    id_num |= class_ << 6
    id_num |= method << 5

    if tag >= 31:
        cont_bit = 0
        while tag > 0:
            header = chr_cls(cont_bit | (tag & 0x7f)) + header
            if not cont_bit:
                cont_bit = 0x80
            tag = tag >> 7
        header = chr_cls(id_num | 31) + header
    else:
        header += chr_cls(id_num | tag)

    length = len(contents)
    if length <= 127:
        header += chr_cls(length)
    else:
        length_bytes = int_to_bytes(length)
        header += chr_cls(0x80 | len(length_bytes))
        header += length_bytes

    return header



import os
import json
import unittest
import jc.parsers.sfdisk

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-l.out'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_l = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-l-multi.out'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_l_multi = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-d.out'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_d = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-d-multi.out'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_d_multi = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-luB.out'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_luB = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-luM.out'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_luM = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-luS.out'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_luS = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-8/sfdisk-l.out'), 'r', encoding='utf-8') as f:
        centos_8_sfdisk_l = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-8/sfdisk-F.out'), 'r', encoding='utf-8') as f:
        centos_8_sfdisk_F = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/debian10/sfdisk-l.out'), 'r', encoding='utf-8') as f:
        debian_10_sfdisk_l = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/debian10/sfdisk-l2.out'), 'r', encoding='utf-8') as f:
        debian_10_sfdisk_l2 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/debian10/sfdisk-l3.out'), 'r', encoding='utf-8') as f:
        debian_10_sfdisk_l3 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/debian10/sfdisk-F.out'), 'r', encoding='utf-8') as f:
        debian_10_sfdisk_F = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/debian10/sfdisk-F2.out'), 'r', encoding='utf-8') as f:
        debian_10_sfdisk_F2 = f.read()


    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-l.json'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_l_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-l-multi.json'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_l_multi_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-d.json'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_d_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-d-multi.json'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_d_multi_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-luB.json'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_luB_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-luM.json'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_luM_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/sfdisk-luS.json'), 'r', encoding='utf-8') as f:
        centos_7_7_sfdisk_luS_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-8/sfdisk-l.json'), 'r', encoding='utf-8') as f:
        centos_8_sfdisk_l_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-8/sfdisk-F.json'), 'r', encoding='utf-8') as f:
        centos_8_sfdisk_F_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/debian10/sfdisk-l.json'), 'r', encoding='utf-8') as f:
        debian_10_sfdisk_l_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/debian10/sfdisk-l2.json'), 'r', encoding='utf-8') as f:
        debian_10_sfdisk_l2_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/debian10/sfdisk-l3.json'), 'r', encoding='utf-8') as f:
        debian_10_sfdisk_l3_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/debian10/sfdisk-F.json'), 'r', encoding='utf-8') as f:
        debian_10_sfdisk_F_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/debian10/sfdisk-F2.json'), 'r', encoding='utf-8') as f:
        debian_10_sfdisk_F2_json = json.loads(f.read())


    def test_sfdisk_nodata(self):
        """
        Test 'sfdisk' with no data
        """
        self.assertEqual(jc.parsers.sfdisk.parse('', quiet=True), [])

    def test_sfdisk_l_centos_7_7(self):
        """
        Test 'sfdisk -l' on Centos 7.7
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.centos_7_7_sfdisk_l, quiet=True), self.centos_7_7_sfdisk_l_json)

    def test_sfdisk_l_multi_centos_7_7(self):
        """
        Test 'sfdisk -l' with multiple disk data on Centos 7.7
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.centos_7_7_sfdisk_l_multi, quiet=True), self.centos_7_7_sfdisk_l_multi_json)

    def test_sfdisk_d_centos_7_7(self):
        """
        Test 'sfdisk -d' on Centos 7.7
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.centos_7_7_sfdisk_d, quiet=True), self.centos_7_7_sfdisk_d_json)

    def test_sfdisk_d_multi_centos_7_7(self):
        """
        Test 'sfdisk -d' with multiple disk data on Centos 7.7
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.centos_7_7_sfdisk_d_multi, quiet=True), self.centos_7_7_sfdisk_d_multi_json)

    def test_sfdisk_luB_centos_7_7(self):
        """
        Test 'sfdisk -luB' on Centos 7.7
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.centos_7_7_sfdisk_luB, quiet=True), self.centos_7_7_sfdisk_luB_json)

    def test_sfdisk_luM_centos_7_7(self):
        """
        Test 'sfdisk -luM' on Centos 7.7
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.centos_7_7_sfdisk_luM, quiet=True), self.centos_7_7_sfdisk_luM_json)

    def test_sfdisk_luS_centos_7_7(self):
        """
        Test 'sfdisk -luS' on Centos 7.7
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.centos_7_7_sfdisk_luS, quiet=True), self.centos_7_7_sfdisk_luS_json)

    def test_sfdisk_l_centos_8(self):
        """
        Test 'sfdisk -l' on Centos 8
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.centos_8_sfdisk_l, quiet=True), self.centos_8_sfdisk_l_json)

    def test_sfdisk_F_centos_8(self):
        """
        Test 'sfdisk -F' on Centos 8
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.centos_8_sfdisk_F, quiet=True), self.centos_8_sfdisk_F_json)

    def test_sfdisk_l_debian_10(self):
        """
        Test 'sfdisk -l' on Debian 10
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.debian_10_sfdisk_l, quiet=True), self.debian_10_sfdisk_l_json)

    def test_sfdisk_l2_debian_10(self):
        """
        Test 'sfdisk -l' on Debian 10 (second example)
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.debian_10_sfdisk_l2, quiet=True), self.debian_10_sfdisk_l2_json)

    def test_sfdisk_l3_debian_10(self):
        """
        Test 'sfdisk -l' on Debian 10 (third example)
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.debian_10_sfdisk_l3, quiet=True), self.debian_10_sfdisk_l3_json)

    def test_sfdisk_F_debian10(self):
        """
        Test 'sfdisk -F' on Debian 10
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.debian_10_sfdisk_F, quiet=True), self.debian_10_sfdisk_F_json)

    def test_sfdisk_F2_debian10(self):
        """
        Test 'sfdisk -F' on Debian 10 (second example)
        """
        self.assertEqual(jc.parsers.sfdisk.parse(self.debian_10_sfdisk_F2, quiet=True), self.debian_10_sfdisk_F2_json)


if __name__ == '__main__':
    unittest.main()



r"""jc - JSON Convert `csv` implicit header file parser

The `csv` implicit header parser will attempt to automatically detect the
delimiter character. If the delimiter cannot be detected it will default to
comma. The file must have no header, and the field names will be generated
as "c0", "c1", etc.

Usage (cli):

    $ cat file.csv | jc --csv-ih

Usage (module):

    import jc
    result = jc.parse('csv_ih', csv_output)

Schema:

CSV file converted to a Dictionary:
https://docs.python.org/3/library/csv.html

    [
      {
        "column_name1":     string,
        "column_name2":     string
      }
    ]

Examples:

    $ cat homes.csv
    142, 160, 28, 10, 5, 3,  60, 0.28,  3167
    175, 180, 18,  8, 4, 1,  12, 0.43,  4033
    129, 132, 13,  6, 3, 1,  41, 0.33,  1471
    ...

    $ cat homes.csv | jc --csv-ih -p
    [
      {
        "c0": "142",
        "c1": "160",
        "c2": "28",
        "c3": "10",
        "c4": "5",
        "c5": "3",
        "c6": "60",
        "c7": "0.28",
        "c8": "3167"
      },
      {
        "c0": "175",
        "c1": "180",
        "c2": "18",
        "c3": "8",
        "c4": "4",
        "c5": "1",
        "c6": "12",
        "c7": "0.43",
        "c8": "4033"
      },
      {
        "c0": "129",
        "c1": "132",
        "c2": "13",
        "c3": "6",
        "c4": "3",
        "c5": "1",
        "c6": "41",
        "c7": "0.33",
        "c8": "1471"
      },
      ...
    ]
"""
from typing import List, Union
from jc.jc_types import JSONDictType
import jc.parsers.csv as jc_csv
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = 'CSV implicit header file parser'
    author = 'Gary Gurlaskie'
    author_email = 'https://github.com/garyg1'
    details = 'Using the python standard csv library'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    tags = ['standard', 'file', 'string']


__version__ = info.version

def parse(
    data: Union[str, bytes],
    raw: bool = False,
    quiet: bool = False
) -> List[JSONDictType]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)
    return jc_csv.parse(data, raw, quiet, implicit_header=True)


import os
import unittest
from jc.exceptions import ParseError
import jc.parsers.asciitable

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    def test_asciitable_nodata(self):
        """
        Test 'asciitable' with no data
        """
        self.assertEqual(jc.parsers.asciitable.parse('', quiet=True), [])

    def test_asciitable_m_pure_ascii(self):
        """
        Test 'asciitable' with a pure ASCII table
        """
        input = '''
+========+========+========+========+========+========+========+
| type   | tota   | used   | fr ee  | shar   | buff   | avai   |

+========+========+========+========+========+========+========+
| Mem    | 3861   | 2228   | 3364   | 1183   | 2743   | 3389   |
+--------+--------+--------+--------+--------+--------+--------+
|        |        |        |        | test 2 |        |        |
+--------+--------+--------+--------+--------+--------+--------+
| last   | last   | last   | ab cde |        |        | final  |
+========+========+========+========+========+========+========+
        '''
        expected = [
            {
                "type": "Mem",
                "tota": "3861",
                "used": "2228",
                "fr_ee": "3364",
                "shar": "1183",
                "buff": "2743",
                "avai": "3389"
            },
            {
                "type": None,
                "tota": None,
                "used": None,
                "fr_ee": None,
                "shar": "test 2",
                "buff": None,
                "avai": None
            },
            {
                "type": "last",
                "tota": "last",
                "used": "last",
                "fr_ee": "ab cde",
                "shar": None,
                "buff": None,
                "avai": "final"
            }
        ]

        self.assertEqual(jc.parsers.asciitable.parse(input, quiet=True), expected)

    def test_asciitable_m_unicode(self):
        """
        Test 'asciitable' with a unicode table
        """
        input = '''
╒════════╤════════╤════════╤════════╤════════╤════════╤════════╕
│ type   │ total  │ used   │ fr ee  │ shar   │ buff   │ avai   │
╞════════╪════════╪════════╪════════╪════════╪════════╪════════╡
│ Mem    │ 3861   │ 2228   │ 3364   │ 1183   │ 2743   │ 3389   │
├────────┼────────┼────────┼────────┼────────┼────────┼────────┤
│ Swap   │ 2097   │ 0      │ 2097   │        │        │        │
├────────┼────────┼────────┼────────┼────────┼────────┼────────┤
│ last   │ last   │ last   │ ab cde │        │        │ final  │
╘════════╧════════╧════════╧════════╧════════╧════════╧════════╛
        '''
        expected = [
            {
                "type": "Mem",
                "total": "3861",
                "used": "2228",
                "fr_ee": "3364",
                "shar": "1183",
                "buff": "2743",
                "avai": "3389"
            },
            {
                "type": "Swap",
                "total": "2097",
                "used": "0",
                "fr_ee": "2097",
                "shar": None,
                "buff": None,
                "avai": None
            },
            {
                "type": "last",
                "total": "last",
                "used": "last",
                "fr_ee": "ab cde",
                "shar": None,
                "buff": None,
                "avai": "final"
            }
        ]

        self.assertEqual(jc.parsers.asciitable.parse(input, quiet=True), expected)

    def test_asciitable_pure_ascii_extra_spaces(self):
        """
        Test 'asciitable' with a pure ASCII table that has heading and
        trailing spaces and newlines.
        """
        input = '''
    
      
    +========+========+========+========+========+========+========+
    | type   | total  | used   | fr ee  | shar   | buff   | avai        
    +========+========+========+========+========+========+========+   
    | Mem    | 3861   | 2228   | 3364   | 1183   | 2743   | 3389     
    +--------+--------+--------+--------+--------+--------+--------+
             |        |        |        | test 2 |        |
    +--------+--------+--------+--------+--------+--------+--------+       
    | last   | last   | last   | ab cde |        |        | final  |
    +========+========+========+========+========+========+========+    
     
  
        '''
        expected = [
            {
                "type": "Mem",
                "total": "3861",
                "used": "2228",
                "fr_ee": "3364",
                "shar": "1183",
                "buff": "2743",
                "avai": "3389"
            },
            {
                "type": None,
                "total": None,
                "used": None,
                "fr_ee": None,
                "shar": "test 2",
                "buff": None,
                "avai": None
            },
            {
                "type": "last",
                "total": "last",
                "used": "last",
                "fr_ee": "ab cde",
                "shar": None,
                "buff": None,
                "avai": "final"
            }
        ]

        self.assertEqual(jc.parsers.asciitable.parse(input, quiet=True), expected)

    def test_asciitable_unicode_extra_spaces(self):
        """
        Test 'asciitable' with a pure ASCII table that has heading and
        trailing spaces and newlines.
        """
        input = '''
    
  
        ╒════════╤════════╤════════╤════════╤════════╤════════╤════════╕
          type   │ total  │ used   │ free   │ shar   │ buff   │ avai
        ╞════════╪════════╪════════╪════════╪════════╪════════╪════════╡      
          Mem    │ 3861   │ 2228   │ 3364   │ 1183   │ 2743   │ 3389     
        ├────────┼────────┼────────┼────────┼────────┼────────┼────────┤  
          Swap   │ 2097   │ 0      │ 2097   │        │        │
        ╘════════╧════════╧════════╧════════╧════════╧════════╧════════╛
   
 
        '''
        expected = [
            {
                "type": "Mem",
                "total": "3861",
                "used": "2228",
                "free": "3364",
                "shar": "1183",
                "buff": "2743",
                "avai": "3389"
            },
            {
                "type": "Swap",
                "total": "2097",
                "used": "0",
                "free": "2097",
                "shar": None,
                "buff": None,
                "avai": None
            }
        ]

        self.assertEqual(jc.parsers.asciitable.parse(input, quiet=True), expected)

    def test_asciitable_markdown(self):
        """
        Test 'asciitable' with a markdown table
        """
        input = '''
        | type   |   total |   used |    free |   shared |   buff cache |   available |
        |--------|---------|--------|---------|----------|--------------|-------------|
        | Mem    | 3861332 | 222820 | 3364176 |    11832 |       274336 |     3389588 |
        | Swap   | 2097148 |      0 | 2097148 |          |              |             |
        '''
        
        expected = [
            {
                "type": "Mem",
                "total": "3861332",
                "used": "222820",
                "free": "3364176",
                "shared": "11832",
                "buff_cache": "274336",
                "available": "3389588"
            },
            {
                "type": "Swap",
                "total": "2097148",
                "used": "0",
                "free": "2097148",
                "shared": None,
                "buff_cache": None,
                "available": None
            }
        ]

        self.assertEqual(jc.parsers.asciitable.parse(input, quiet=True), expected)

    def test_asciitable_simple(self):
        """
        Test 'asciitable' with a simple table
        """
        input = '''
        type      total    used     free    shared    buff cache    available
        ------  -------  ------  -------  --------  ------------  -----------
        Mem     3861332  222820  3364176     11832        274336      3389588
        Swap    2097148       0  2097148
        '''

        expected = [
            {
                "type": "Mem",
                "total": "3861332",
                "used": "222820",
                "free": "3364176",
                "shared": "11832",
                "buff_cache": "274336",
                "available": "3389588"
            },
            {
                "type": "Swap",
                "total": "2097148",
                "used": "0",
                "free": "2097148",
                "shared": None,
                "buff_cache": None,
                "available": None
            }
        ]

        self.assertEqual(jc.parsers.asciitable.parse(input, quiet=True), expected)

    def test_asciitable_pretty_ansi(self):
        """
        Test 'asciitable' with a pretty table with ANSI codes
        """
        input = '''┏━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━┓                                   \n                                   ┃\x1b[1m \x1b[0m\x1b[1mReleased    \x1b[0m\x1b[1m \x1b[0m┃\x1b[1m \x1b[0m\x1b[1mTitle                            \x1b[0m\x1b[1m \x1b[0m┃\x1b[1m \x1b[0m\x1b[1m    Box Office\x1b[0m\x1b[1m \x1b[0m┃                                   \n                                   ┡━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━┩                                   \n                                   │\x1b[36m \x1b[0m\x1b[36mDec 20, 2019\x1b[0m\x1b[36m \x1b[0m│\x1b[35m \x1b[0m\x1b[35mStar Wars: The Rise of Skywalker \x1b[0m\x1b[35m \x1b[0m│\x1b[32m \x1b[0m\x1b[32m  $952,110,690\x1b[0m\x1b[32m \x1b[0m│                                   \n                                   │\x1b[36m \x1b[0m\x1b[36mMay 25, 2018\x1b[0m\x1b[36m \x1b[0m│\x1b[35m \x1b[0m\x1b[35mSolo: A Star Wars Story          \x1b[0m\x1b[35m \x1b[0m│\x1b[32m \x1b[0m\x1b[32m  $393,151,347\x1b[0m\x1b[32m \x1b[0m│                                   \n                                   │\x1b[36m \x1b[0m\x1b[36mDec 15, 2017\x1b[0m\x1b[36m \x1b[0m│\x1b[35m \x1b[0m\x1b[35mStar Wars Ep. V111: The Last Jedi\x1b[0m\x1b[35m \x1b[0m│\x1b[32m \x1b[0m\x1b[32m$1,332,539,889\x1b[0m\x1b[32m \x1b[0m│                                   \n                                   │\x1b[36m \x1b[0m\x1b[36mDec 16, 2016\x1b[0m\x1b[36m \x1b[0m│\x1b[35m \x1b[0m\x1b[35mRogue One: A Star Wars Story     \x1b[0m\x1b[35m \x1b[0m│\x1b[32m \x1b[0m\x1b[32m$1,332,439,889\x1b[0m\x1b[32m \x1b[0m│                                   \n                                   └──────────────┴───────────────────────────────────┴────────────────┘                                   \n'''

        expected = [
            {
                "released": "Dec 20, 2019",
                "title": "Star Wars: The Rise of Skywalker",
                "box_office": "$952,110,690"
            },
            {
                "released": "May 25, 2018",
                "title": "Solo: A Star Wars Story",
                "box_office": "$393,151,347"
            },
            {
                "released": "Dec 15, 2017",
                "title": "Star Wars Ep. V111: The Last Jedi",
                "box_office": "$1,332,539,889"
            },
            {
                "released": "Dec 16, 2016",
                "title": "Rogue One: A Star Wars Story",
                "box_office": "$1,332,439,889"
            }
        ]

        self.assertEqual(jc.parsers.asciitable.parse(input, quiet=True), expected)

    def test_asciitable_special_chars_in_header(self):
        """
        Test 'asciitable' with a pure ASCII table that has special
        characters in the header. These should be converted to underscores
        and no trailing or consecutive underscores should end up in the
        resulting key names.
        """
        input = '''
Protocol  Address     Age (min)  Hardware Addr   Type   Interface
Internet  10.12.13.1        98   0950.5785.5cd1  ARPA   FastEthernet2.13
Internet  10.12.13.3       131   0150.7685.14d5  ARPA   GigabitEthernet2.13
Internet  10.12.13.4       198   0950.5C8A.5c41  ARPA   GigabitEthernet2.17
        '''

        expected = [
            {
                "protocol": "Internet",
                "address": "10.12.13.1",
                "age_min": "98",
                "hardware_addr": "0950.5785.5cd1",
                "type": "ARPA",
                "interface": "FastEthernet2.13"
            },
            {
                "protocol": "Internet",
                "address": "10.12.13.3",
                "age_min": "131",
                "hardware_addr": "0150.7685.14d5",
                "type": "ARPA",
                "interface": "GigabitEthernet2.13"
            },
            {
                "protocol": "Internet",
                "address": "10.12.13.4",
                "age_min": "198",
                "hardware_addr": "0950.5C8A.5c41",
                "type": "ARPA",
                "interface": "GigabitEthernet2.17"
            }
        ]

        self.assertEqual(jc.parsers.asciitable.parse(input, quiet=True), expected)

    def test_asciitable_no_lower_raw(self):
        """
        Test 'asciitable' with a pure ASCII table that has special
        characters and mixed case in the header. These should be converted to underscores
        and no trailing or consecutive underscores should end up in the
        resulting key names. Using `raw` in this test to preserve case. (no lower)
        """
        input = '''
Protocol  Address     Age (min)  Hardware Addr   Type   Interface
Internet  10.12.13.1        98   0950.5785.5cd1  ARPA   FastEthernet2.13
Internet  10.12.13.3       131   0150.7685.14d5  ARPA   GigabitEthernet2.13
Internet  10.12.13.4       198   0950.5C8A.5c41  ARPA   GigabitEthernet2.17
        '''

        expected = [
            {
                "Protocol": "Internet",
                "Address": "10.12.13.1",
                "Age_min": "98",
                "Hardware_Addr": "0950.5785.5cd1",
                "Type": "ARPA",
                "Interface": "FastEthernet2.13"
            },
            {
                "Protocol": "Internet",
                "Address": "10.12.13.3",
                "Age_min": "131",
                "Hardware_Addr": "0150.7685.14d5",
                "Type": "ARPA",
                "Interface": "GigabitEthernet2.13"
            },
            {
                "Protocol": "Internet",
                "Address": "10.12.13.4",
                "Age_min": "198",
                "Hardware_Addr": "0950.5C8A.5c41",
                "Type": "ARPA",
                "Interface": "GigabitEthernet2.17"
            }
        ]

        self.assertEqual(jc.parsers.asciitable.parse(input, raw=True, quiet=True), expected)

    def test_asciitable_centered_col_header(self):
        """
        Test 'asciitable' with long centered column header which can break
        column alignment
        """
        input = '''
            +---------+--------------------------+-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------+------------------+
            | fdc_id  |        data_type         |                                                                                                                           description                                                                                                                           | food_category_id | publication_date |
            +---------+--------------------------+-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------+------------------+
            | 167512  | sr_legacy_food           | Pillsbury Golden Layer Buttermilk Biscuits, Artificial Flavor, refrigerated dough                                                                                                                                                                               |                  | 2019-04-01       |
            | 167513  | sr_legacy_food           | Pillsbury, Cinnamon Rolls with Icing, refrigerated dough                                                                                                                                                                                                        |                  | 2019-04-01       |
            | 167514  | sr_legacy_food           | Kraft Foods, Shake N Bake Original Recipe, Coating for Pork, dry                                                                                                                                                                                                |                  | 2019-04-01       |
            | 167515  | sr_legacy_food           | George Weston Bakeries, Thomas English Muffins                                                                                                                                                                                                                  |                  | 2019-04-01       |
            | 167516  | sr_legacy_food           | Waffles, buttermilk, frozen, ready-to-heat                                                                                                                                                                                                                      |                  | 2019-04-01       |
            | 167517  | sr_legacy_food           | Waffle, buttermilk, frozen, ready-to-heat, toasted                                                                                                                                                                                                              |                  | 2019-04-01       |
            | 167518  | sr_legacy_food           | Waffle, buttermilk, frozen, ready-to-heat, microwaved                                                                                                                                                                                                           |                  | 2019-04-01       |
        '''

        expected = [
            {
                "fdc_id": "167512",
                "data_type": "sr_legacy_food",
                "description": "Pillsbury Golden Layer Buttermilk Biscuits, Artificial Flavor, refrigerated dough",
                "food_category_id": None,
                "publication_date": "2019-04-01"
            },
            {
                "fdc_id": "167513",
                "data_type": "sr_legacy_food",
                "description": "Pillsbury, Cinnamon Rolls with Icing, refrigerated dough",
                "food_category_id": None,
                "publication_date": "2019-04-01"
            },
            {
                "fdc_id": "167514",
                "data_type": "sr_legacy_food",
                "description": "Kraft Foods, Shake N Bake Original Recipe, Coating for Pork, dry",
                "food_category_id": None,
                "publication_date": "2019-04-01"
            },
            {
                "fdc_id": "167515",
                "data_type": "sr_legacy_food",
                "description": "George Weston Bakeries, Thomas English Muffins",
                "food_category_id": None,
                "publication_date": "2019-04-01"
            },
            {
                "fdc_id": "167516",
                "data_type": "sr_legacy_food",
                "description": "Waffles, buttermilk, frozen, ready-to-heat",
                "food_category_id": None,
                "publication_date": "2019-04-01"
            },
            {
                "fdc_id": "167517",
                "data_type": "sr_legacy_food",
                "description": "Waffle, buttermilk, frozen, ready-to-heat, toasted",
                "food_category_id": None,
                "publication_date": "2019-04-01"
            },
            {
                "fdc_id": "167518",
                "data_type": "sr_legacy_food",
                "description": "Waffle, buttermilk, frozen, ready-to-heat, microwaved",
                "food_category_id": None,
                "publication_date": "2019-04-01"
            }
        ]

        self.assertEqual(jc.parsers.asciitable.parse(input, quiet=True), expected)


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import json
import jc.parsers.jar_manifest

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/rhel-8/MANIFEST.MF.out'), 'r', encoding='utf-8') as f:
        rhel_8_manifest_mf = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/rhel-8/MANIFEST.MF.MULTI.out'), 'r', encoding='utf-8') as f:
        rhel_8_manifest_mf_multi = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/rhel-8/MANIFEST.MF.json'), 'r', encoding='utf-8') as f:
        rhel_8_manifest_mf_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/rhel-8/MANIFEST.MF.MULTI.json'), 'r', encoding='utf-8') as f:
        rhel_8_manifest_mf_multi_json = json.loads(f.read())


    def test_jar_manifest_nodata(self):
        """
        Test 'jar_manifest' parser with no data
        """
        self.assertEqual(jc.parsers.jar_manifest.parse('', quiet=True), [])

    def test_jar_manifest_rhel_8(self):
        """
        Test 'cat MANIFEST.MF | jc --jar_manifest'
        """
        self.assertEqual(jc.parsers.jar_manifest.parse(self.rhel_8_manifest_mf, quiet=True), self.rhel_8_manifest_mf_json)

    def test_jar_manifest_multi_rhel_8(self):
        """
        Test 'unzip -c apache-log4j-2.16.0-bin/log4j-core-2.16.0.jar META-INF/MANIFEST.MF | jc --jar_manifest'
        """
        self.assertEqual(jc.parsers.jar_manifest.parse(self.rhel_8_manifest_mf_multi, quiet=True), self.rhel_8_manifest_mf_multi_json)


if __name__ == '__main__':
    unittest.main()
